#
#  shared_enum.py -
#
#  Created:   February 21, 2020
#  Updated:
#
#  Author:    See AUTHORS
#

from aenum import Enum, IntEnum


class UnitSystem(Enum):
    """SWMM Unit System enum class. 

    .. rubric:: Enum Members

    ================ ====================
    :attr:`~US`      US Traditional units
    :attr:`~SI`      SI (metric) units
    ================ ====================
    """
    US = 0
    SI = 1


class FlowUnits(Enum):
    """SWMM Flow Units enum class.

    .. rubric:: Enum Members

    ================ =======================
    :attr:`~CFS`     cubic feet per second
    :attr:`~GPM`     gallons per minute
    :attr:`~MGD`     million gallons per day
    :attr:`~CMS`     cubic meters per second
    :attr:`~LPS`     liters per second
    :attr:`~MLD`     million liters per day
    ================ =======================
    """
    CFS = 0
    GPM = 1
    MGD = 2
    CMS = 3
    LPS = 4
    MLD = 5


class ConcUnits(Enum):
    """SWMM Concentration Units enum class.

    .. rubric:: Enum Members

    ================ ====================
    :attr:`~MG`      Milligrams per liter
    :attr:`~UG`      Micrograms per liter
    :attr:`~COUNT`   Counts per liter
    :attr:`~NONE`    None
    ================ ====================
    """
    MG = 0
    UG = 1
    COUNT = 2
    NONE = 3


class BaseUnits(Enum, start = 1):
    """Base Units enum class

    .. rubric:: Enum Members

    =================== ==================
    :attr:`~RAIN_INT`   rainfall intensity
    :attr:`~SNOW_DEPTH` rainfall depth  
    :attr:`~EVAP_RATE`  evaporation rate
    :attr:`~INFIL_RATE` infiltration rate
    :attr:`~FLOW_RATE`  flow rate
    :attr:`~ELEV`       elevation
    :attr:`~PERCENT`    percent
    :attr:`~CONCEN`     concentration
    :attr:`~HEAD`       head
    :attr:`~VOLUME`     volume
    :attr:`~VELOCITY`   velocity
    :attr:`~TEMP`       temperature
    :attr:`~UNITLESS`   unitless quantity
    :attr:`~NONE`       none
    =================== ==================
    """
    RAIN_INT = 1
    SNOW_DEPTH = 2
    EVAP_RATE = 3
    INFIL_RATE = 4
    FLOW_RATE = 5
    ELEV = 6
    PERCENT = 7
    CONCEN = 8
    HEAD = 9
    VOLUME = 10
    VELOCITY = 11
    TEMP = 12
    UNITLESS = 13
    NONE = 14


class ObjectType(Enum):
    """SWMM Object Types enum class.

    .. rubric:: Enum Members

    ==================== ===============================
    :attr:`~GAGE`        rain gage 
    :attr:`~SUBCATCH`    subcatchment
    :attr:`~NODE`        conveyance system node
    :attr:`~LINK`        conveyance system link
    :attr:`~POLLUT`      pollutant
    :attr:`~LANDUSE`     land use category
    :attr:`~TIMEPATTERN` dry weather flow time pattern
    :attr:`~CURVE`       generic table of values
    :attr:`~TSERIES`     generic time series of values
    :attr:`~CONTROL`     conveyance system control rules
    :attr:`~TRANSECT`    irregular channel cross-section
    :attr:`~AQUIFER`     groundwater aquifer
    :attr:`~UNITHYD`     RDII unit hydrograph
    :attr:`~SNOWMELT`    snowmelt parameter set
    :attr:`~SHAPE`       custom conduit shape
    :attr:`~LID`         LID treatment unit
    ==================== ===============================
    """
    GAGE = 0
    SUBCATCH = 1
    NODE = 2
    LINK = 3
    POLLUT = 4
    LANDUSE = 5
    TIMEPATTERN = 6
    CURVE = 7
    TSERIES = 8
    CONTROL = 9
    TRANSECT = 10
    AQUIFER = 11
    UNITHYD = 12
    SNOWMELT = 13
    SHAPE = 14
    LID = 15


class NodeType(Enum):
    """Node Sub-Types enum class.

    :attr:`~JUNCTION`
    :attr:`~OUTFALL`
    :attr:`~STORAGE`
    :attr:`~DIVIDER`
    """
    JUNCTION = 0
    OUTFALL = 1
    STORAGE = 2
    DIVIDER = 3


class LinkType(Enum):
    """Link Sub-Types enum class.

    :attr:`~CONDUIT`
    :attr:`~PUMP`
    :attr:`~ORIFICE`
    :attr:`~WEIR`
    :attr:`~OUTLET`
    """
    CONDUIT = 0
    PUMP = 1
    ORIFICE = 2
    WEIR = 3
    OUTLET = 4


class LinkDirection(Enum):
    """Link Direction enum class.

    :attr:`~UPSTREAM_TO_DOWNSTREAM`
    :attr:`~DOWNSTREAM_TO_UPSTREAM`
    """
    UPSTREAM_TO_DOWNSTREAM = 1
    DOWNSTREAM_TO_UPSTREAM = -1
    

class ElementType(IntEnum, start = 0):
    """SWMM Element Types enum class.

    :attr:`~SUBCATCH`
    :attr:`~NODE`
    :attr:`~LINK`
    :attr:`~SYSTEM`
    :attr:`~POLLUT`
    """
    SUBCATCH = 0
    NODE = 1
    LINK = 2
    SYSTEM = 3
    POLLUT = 4


class Time(Enum, start = 0):
    """Time enum class.

    :attr:`~REPORT_STEP`
    :attr:`~NUM_PERIODS`
    """
    REPORT_STEP = 0
    NUM_PERIODS = 1


class SubcatchAttribute(Enum, start = 0):
    """Subcatchment Attributes enum class.

    .. rubric:: Enum Members

    ========================= ===============================
    :attr:`~RAINFALL`         rainfall intensity
    :attr:`~SNOW_DEPTH`       snow depth
    :attr:`~EVAP_LOSS`        evaporation loss 
    :attr:`~INFIL_LOSS`       infiltration loss
    :attr:`~RUNOFF_RATE`      runoff flow rate
    :attr:`~GW_OUTFLOW_RATE`  groundwater flow rate to node
    :attr:`~GW_TABLE_ELEV`    elevation of saturated gw table
    :attr:`~SOIL_MOISTURE`    soil moisture
    :attr:`~POLLUT_CONC_0`    pollutant washoff concentration
    ========================= ===============================
    """    
    RAINFALL = 0
    SNOW_DEPTH = 1
    EVAP_LOSS = 2
    INFIL_LOSS = 3
    RUNOFF_RATE = 4
    GW_OUTFLOW_RATE = 5
    GW_TABLE_ELEV = 6
    SOIL_MOISTURE = 7
    POLLUT_CONC_0 = 8


class NodeAttribute(Enum, start = 0):
    """Node Attribute enum class.

    .. rubric:: Enum Members

    ========================= ===============================
    :attr:`~INVERT_DEPTH`     water depth above invert
    :attr:`~HYDRAULIC_HEAD`   hydraulic head
    :attr:`~PONDED_VOLUME`    volume stored and ponded
    :attr:`~LATERAL_INFLOW`   lateral inflow rate
    :attr:`~TOTAL_INFLOW`     total inflow rate
    :attr:`~FLOODING_LOSSES`  overflow rate
    :attr:`~POLLUT_CONC_0`    concentration of each pollutant
    ========================= ===============================
    """
    INVERT_DEPTH = 0
    HYDRAULIC_HEAD = 1
    PONDED_VOLUME = 2
    LATERAL_INFLOW = 3
    TOTAL_INFLOW = 4
    FLOODING_LOSSES = 5
    POLLUT_CONC_0 = 6


class LinkAttribute(Enum, start = 0):
    """Link Attribute enum class.

    .. rubric:: Enum Members

    ====================== ===============================
    :attr:`~FLOW_RATE`     flow rate
    :attr:`~FLOW_DEPTH`    flow depth
    :attr:`~FLOW_VELOCITY` flow velocity
    :attr:`~FLOW_VOLUME`   link volume
    :attr:`~CAPACITY`      ratio of area to full area
    :attr:`~POLLUT_CONC_0` concentration of each pollutant
    ====================== ===============================
    """
    FLOW_RATE = 0
    FLOW_DEPTH = 1
    FLOW_VELOCITY = 2
    FLOW_VOLUME = 3
    CAPACITY = 4
    POLLUT_CONC_0 = 5


class SystemAttribute(Enum, start = 0):
    """System Attribute enum class.

    .. rubric:: Enum Members

    ============================= ====================
    :attr:`~AIR_TEMP`             air temperature
    :attr:`~RAINFALL`             rainfall intensity
    :attr:`~SNOW_DEPTH`           snow depth
    :attr:`~EVAP_INFIL_LOSS`      infiltration
    :attr:`~RUNOFF_FLOW`          runoff
    :attr:`~DRY_WEATHER_INFLOW`   dry weather inflow
    :attr:`~GW_INFLOW`            ground water inflow
    :attr:`~RDII_INFLOW`          RDII inflow
    :attr:`~DIRECT_INFLOW`        external inflow
    :attr:`~TOTAL_LATERAL_INFLOW` total lateral inflow
    :attr:`~FLOOD_LOSSES`         flooding outflow
    :attr:`~OUTFALL_FLOWS`        outfall outflow
    :attr:`~VOLUME_STORED`        storage volume
    :attr:`~EVAP_RATE`            evaporation
    :attr:`~PTNL_EVAP_RATE`       potential evapotranspiration
    ============================= ====================
    """
    AIR_TEMP = 0
    RAINFALL = 1
    SNOW_DEPTH = 2
    EVAP_INFIL_LOSS = 3
    RUNOFF_FLOW = 4
    DRY_WEATHER_INFLOW = 5
    GW_INFLOW = 6
    RDII_INFLOW = 7
    DIRECT_INFLOW = 8
    TOTAL_LATERAL_INFLOW = 9
    FLOOD_LOSSES = 10
    OUTFALL_FLOWS = 11
    VOLUME_STORED = 12
    EVAP_RATE = 13
    PTNL_EVAP_RATE = 14


#
# Solver Toolkit API Enums
#    
class TimeProperty(Enum):
    """Time Property enum class.

    :attr:`~START_DATE`
    :attr:`~END_DATE`
    :attr:`~REPORT_DATE`
    """
    START_DATE = 0
    END_DATE = 1
    REPORT_DATE = 2


class UnitProperty(Enum):
    """Unit Property enum class.

    :attr:`~SYSTEM_UNIT`
    :attr:`~FLOW_UNIT`
    """
    SYSTEM_UNIT = 0
    FLOW_UNIT = 1


class SimOption(Enum):
    """Simulation Option enum class.

    :attr:`~ALLOW_POND`
    :attr:`~SKIP_STEADY`
    :attr:`~IGNORE_RAIN`
    :attr:`~IGNORE_RDII`
    :attr:`~IGNORE_SNOW`
    :attr:`~IGNORE_GW`
    :attr:`~IGNORE_ROUTE`
    :attr:`~IGNORE_ROUTE_QUALITY`
    """
    ALLOW_POND = 0
    SKIP_STEADY = 1
    IGNORE_RAIN = 2
    IGNORE_RDII = 3
    IGNORE_SNOW = 4
    IGNORE_GW = 5
    IGNORE_ROUTE = 6
    IGNORE_ROUTE_QUALITY = 7


class SimSetting(Enum):
    """Simulation Settings enum class.

    :attr:`~ROUTE_STEP`
    :attr:`~MIN_ROUTE_STEP`
    :attr:`~LENGTH_STEP`
    :attr:`~START_DRY_DAYS`
    :attr:`~COURANT_FACTOR`
    :attr:`~MIN_SURFACE_AREA`
    :attr:`~MIN_SLOPE`
    :attr:`~RUNOFF_ERROR`
    :attr:`~GW_ERROR`
    :attr:`~FLOW_ERROR`
    :attr:`~QUALITY_ERROR`
    :attr:`~HEAD_TOLERANCE`
    :attr:`~SYSTEM_FLOW_TOLERANCE`
    :attr:`~LATERAL_FLOW_TOLERANCE`
    """
    ROUTE_STEP = 0
    MIN_ROUTE_STEP = 1
    LENGTH_STEP = 2
    START_DRY_DAYS = 3
    COURANT_FACTOR = 4
    MIN_SURFACE_AREA = 5
    MIN_SLOPE = 6
    RUNOFF_ERROR = 7
    GW_ERROR = 8
    FLOW_ERROR = 9
    QUALITY_ERROR = 10
    HEAD_TOLERANCE = 11
    SYSTEM_FLOW_TOLERANCE = 12
    LATERAL_FLOW_TOLERANCE = 13


class NodeProperty(Enum):
    """Node Property enum class.

    :attr:`~INVERT_ELEVATION`
    :attr:`~FULL_DEPTH`
    :attr:`~SURCHARGE_DEPTH`
    :attr:`~POND_AREA`
    :attr:`~INITIAL_DEPTH`
    """
    INVERT_ELEVATION = 0
    FULL_DEPTH = 1
    SURCHARGE_DEPTH = 2
    POND_AREA = 3
    INITIAL_DEPTH = 4


class NodeResult(Enum):
    """Node Result enum class.

    :attr:`~TOTAL_INFLOW`
    :attr:`~TOTAL_OUTFLOW`
    :attr:`~LOSSES`
    :attr:`~VOLUME`
    :attr:`~FLOOD`
    :attr:`~DEPTH`
    :attr:`~HEAD`
    :attr:`~LATERAL_INFLOW`
    :attr:`~HYD_RES_TIME`
    """
    TOTAL_INFLOW = 0
    TOTAL_OUTFLOW = 1
    LOSSES = 2
    VOLUME = 3
    FLOOD = 4
    DEPTH = 5
    HEAD = 6
    LATERAL_INFLOW = 7
    HYD_RES_TIME = 8


class NodePollutant(Enum):
    """Node Pollutant enum class.

    :attr:`~QUALITY`
    :attr:`~INFLOW_CONC`
    :attr:`~REACTOR_CONC`
    """
    QUALITY = 0
    INFLOW_CONC = 1
    REACTOR_CONC = 2


class LinkProperty(Enum):
    """Link Property enum class.

    :attr:`~OFFSET_1`
    :attr:`~OFFSET_2`
    :attr:`~INITIAL_FLOW`
    :attr:`~FLOW_LIMIT`
    :attr:`~INLET_LOSS`
    :attr:`~OUTLET_LOSS`
    :attr:`~AVERAGE_LOSS`
    """
    OFFSET_1 = 0
    OFFSET_2 = 1
    INITIAL_FLOW = 2
    FLOW_LIMIT = 3
    INLET_LOSS = 4
    OUTLET_LOSS = 5
    AVERAGE_LOSS = 6


class LinkResult(Enum):
    """Link Result enum class.

    :attr:`~FLOW`
    :attr:`~DEPTH`
    :attr:`~VOLUME`
    :attr:`~US_SURFACE_AREA`
    :attr:`~DS_SURFACE_AREA`
    :attr:`~SETTING`
    :attr:`~TARGET_SETTING`
    :attr:`~FROUDE`
    """
    FLOW = 0
    DEPTH = 1
    VOLUME = 2
    US_SURFACE_AREA = 3
    DS_SURFACE_AREA = 4
    SETTING = 5
    TARGET_SETTING = 6
    FROUDE = 7


class LinkPollutant(Enum):
    """Link Pollutant enum class.

    :attr:`~QUALITY`
    :attr:`~TOTAL_LOAD`
    :attr:`~REACTOR_CONC`
    """
    QUALITY = 0
    TOTAL_LOAD = 1
    REACTOR_CONC = 2

    
class SubcatchProperty(Enum):
    """Subcatchment Property enum class.

    :attr:`~WIDTH`
    :attr:`~AREA`
    :attr:`~IMPERVIOUS_FRACTION`
    :attr:`~SLOPE`
    :attr:`~CURB_LENGTH`
    """
    WIDTH = 0
    AREA = 1
    IMPERVIOUS_FRACTION = 2
    SLOPE = 3
    CURB_LENGTH = 4


class SubcatchResult(Enum):
    """Subcatchment Result enum class.

    :attr:`~RAIN`
    :attr:`~EVAPORATION`
    :attr:`~INFILTRATION`
    :attr:`~RUNON`
    :attr:`~RUNOFF`
    :attr:`~SNOW`
    """
    RAIN = 0
    EVAPORATION = 1
    INFILTRATION = 2
    RUNON = 3
    RUNOFF = 4
    SNOW = 5


class SubcatchPollutant(Enum):
    """Subcatchment Pollutant enum class.

    :attr:`~BUILD_UP`
    :attr:`~CONCENTRATION`
    :attr:`~QUALITY`
    :attr:`~TOTAL_LOAD`
    """
    BUILD_UP = 0
    CONCENTRATION = 1
    QUALITY = 2
    TOTAL_LOAD = 3

    
class LidUsageProperty(Enum):
    """LID Usage Property enum class.

    :attr:`~UNIT_AREA`
    :attr:`~TOP_WIDTH`
    :attr:`~BOTTOM_WIDTH`
    :attr:`~INITIAL_SATURATION`
    :attr:`~FROM_IMPERVIOUS`
    :attr:`~FROM_PERVIOUS`
    """
    UNIT_AREA = 0
    TOP_WIDTH = 1
    BOTTOM_WIDTH = 2
    INITIAL_SATURATION = 3
    FROM_IMPERVIOUS = 4
    FROM_PERVIOUS = 5


class LidUsageOption(Enum):
    """LID Usage Option enum class.

    :attr:`~INDEX`
    :attr:`~NUMBER`
    :attr:`~TO_PERV`
    :attr:`~DRAIN_SUBCATCH`
    :attr:`~DRAIN_NODE`
    """
    INDEX = 0
    NUMBER = 1
    TO_PERV = 2
    DRAIN_SUBCATCH = 3
    DRAIN_NODE = 4


class LidLayer(Enum):
    """LID Layer enum class.

    :attr:`~SURFACE`
    :attr:`~SOIL`
    :attr:`~STORAGE`
    :attr:`~PAVEMENT`
    :attr:`~DRAIN`
    :attr:`~DRAIN_MAT`
    """
    SURFACE = 0
    SOIL = 1
    STORAGE = 2
    PAVEMENT = 3
    DRAIN = 4
    DRAIN_MAT = 5

    
class LidLayerProperty(Enum):
    """LID Layer Property enum class.

    :attr:`~THICKNESS`
    :attr:`~VOID_FRACTION`
    :attr:`~ROUGHNESS`
    :attr:`~SURFACE_SLOPE`
    :attr:`~SIDE_SLOP`
    :attr:`~ALPHA`
    :attr:`~POROSITY`
    :attr:`~FIELD_CAPACITY`
    :attr:`~WILTING_POINT`
    :attr:`~SUCTION_HEAD`
    :attr:`~K_SATURATION`
    :attr:`~K_SLOPE`
    :attr:`~CLOG_FACTOR`
    :attr:`~IMPERVIOUS_FRACTION`
    :attr:`~DRAIN_COEFFICIENT`
    :attr:`~DRAIN_EXPONENT`
    :attr:`~DRAIN_OFFSET`
    :attr:`~DRAIN_DELAY`
    :attr:`~DRAIN_HEAD_OPEN`
    :attr:`~DRAIN_HEAD_CLOSE`
    :attr:`~DRAIN_CURVE`
    :attr:`~DRAIN_REGEN_DAYS`
    :attr:`~DRAIN_REGEN_DEGREE`
    """
    THICKNESS = 0
    VOID_FRACTION = 1
    ROUGHNESS = 2
    SURFACE_SLOPE = 3
    SIDE_SLOP = 4
    ALPHA = 5
    POROSITY = 6
    FIELD_CAPACITY = 7
    WILTING_POINT = 8
    SUCTION_HEAD = 9
    K_SATURATION = 10
    K_SLOPE = 11
    CLOG_FACTOR = 12
    IMPERVIOUS_FRACTION = 13
    DRAIN_COEFFICIENT = 14
    DRAIN_EXPONENT = 15
    DRAIN_OFFSET = 16
    DRAIN_DELAY = 17
    DRAIN_HEAD_OPEN = 18
    DRAIN_HEAD_CLOSE = 19
    DRAIN_CURVE = 20
    DRAIN_REGEN_DAYS = 21
    DRAIN_REGEN_DEGREE = 22


class LidResult(Enum):
    """LID Result enum class.

    :attr:`~INFLOW`
    :attr:`~EVAPORATION`
    :attr:`~INFILTRATION`
    :attr:`~SURFACE_FLOW`
    :attr:`~DRAIN_FLOW`
    :attr:`~INITIAL_VOLUME`
    :attr:`~FINAL_VOLUME`
    :attr:`~SURFACE_DEPTH`
    :attr:`~PAVEMENT_DEPTH`
    :attr:`~SOIL_MOISTURE`
    :attr:`~STORAGE_DEPTH`
    :attr:`~DRY_TIME`
    :attr:`~OLD_DRAIN_FLOW`
    :attr:`~NEW_DRAIN_FLOW`
    :attr:`~PERVIOUS_AREA`
    :attr:`~FLOW_TO_PERVIOUS_AREA`
    :attr:`~EVAPORATE_RATE`
    :attr:`~NATIVE_INFILTRATION`
    :attr:`~SURFACE_INFLOW`
    :attr:`~SURFACE_INFILTRATION`
    :attr:`~SURFACE_EVAPORATION`
    :attr:`~SURFACE_OUTFLOW`
    :attr:`~PAVEMENT_EVAPORATION`
    :attr:`~PAVEMENT_PERCOLATION`
    :attr:`~SOIL_EVAPORATION`
    :attr:`~SOIL_PERCOLATION`
    :attr:`~STORAGE_INFLOW`
    :attr:`~STORAGE_EXFILTRATION`
    :attr:`~STORAGE_EVAPORATION`
    :attr:`~STORAGE_DRAIN`
    """
    INFLOW = 0
    EVAPORATION = 1
    INFILTRATION = 2
    SURFACE_FLOW = 3
    DRAIN_FLOW = 4
    INITIAL_VOLUME = 5
    FINAL_VOLUME = 6
    SURFACE_DEPTH = 7
    PAVEMENT_DEPTH = 8
    SOIL_MOISTURE = 9
    STORAGE_DEPTH = 10
    DRY_TIME = 11
    OLD_DRAIN_FLOW = 12
    NEW_DRAIN_FLOW = 13
    PERVIOUS_AREA = 14
    FLOW_TO_PERVIOUS_AREA = 15
    EVAPORATE_RATE = 16
    NATIVE_INFILTRATION = 17
    SURFACE_INFLOW = 18
    SURFACE_INFILTRATION = 19
    SURFACE_EVAPORATION = 20
    SURFACE_OUTFLOW = 21
    PAVEMENT_EVAPORATION = 22
    PAVEMENT_PERCOLATION = 23
    SOIL_EVAPORATION = 24
    SOIL_PERCOLATION = 25
    STORAGE_INFLOW = 26
    STORAGE_EXFILTRATION = 27
    STORAGE_EVAPORATION = 28
    STORAGE_DRAIN = 29

    
class RainResult(Enum):
    """Rain Result enum class.

    :attr:`~TOTAL`
    :attr:`~RAINFALL`
    :attr:`~SNOWFALL`
    """
    TOTAL = 0
    RAINFALL = 1
    SNOWFALL = 2



#
#   test_output.py
#
#   Created: Aug 7, 2018
#   Updated: Apr 20, 2020
#
#   Author:  See AUTHORS
#
#   Unit testing for SWMM Output API using pytest.
#


import os

import pytest
import numpy as np

from swmm.toolkit import output, shared_enum, output_metadata


DATA_PATH = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'data')
OUTPUT_FILE_EXAMPLE1 = os.path.join(DATA_PATH, 'test_Example1.out')


def test_openclose():
    _handle = output.init()
    output.open(_handle, OUTPUT_FILE_EXAMPLE1)
    output.close(_handle)


def test_opencloseopenclose():
    _handle = output.init()
    output.open(_handle, OUTPUT_FILE_EXAMPLE1)
    output.close(_handle)
    _handle2 = output.init()
    output.open(_handle2, OUTPUT_FILE_EXAMPLE1)
    output.close(_handle2)


@pytest.fixture()
def handle(request):
    _handle = output.init()
    output.open(_handle, OUTPUT_FILE_EXAMPLE1)

    def close():
        output.close(_handle)

    request.addfinalizer(close)
    return _handle


def test_getversion(handle):

  assert output.get_version(handle) == 51000


def test_getprojectsize(handle):

   assert output.get_proj_size(handle) == [8, 14, 13, 1, 2]


# def test_getflowunits(handle):
#
#    assert output.getunits(handle)[0] == output.FlowUnits.CFS.value
#
#
# def test_getpollutantunits(handle):
#
#    assert output.getunits(handle)[2:] == [output.ConcUnits.MG.value, output.ConcUnits.UG.value]


def test_getstartdate(handle):

    assert output.get_start_date(handle) == 35796


def test_gettimes(handle):

    assert output.get_times(handle, shared_enum.Time.REPORT_STEP) == 3600
    assert output.get_times(handle, shared_enum.Time.NUM_PERIODS) == 36


def test_getelementname(handle):

    assert output.get_elem_name(handle, shared_enum.ElementType.NODE, 1) == "10"


def test_getsubcatchseries(handle):

    ref_array = np.array([0.0,
                          1.2438242,
                          2.5639679,
                          4.524055,
                          2.5115132,
                          0.69808137,
                          0.040894926,
                          0.011605669,
                          0.00509294,
                          0.0027438672])

    test_array = output.get_subcatch_series(handle, 1, shared_enum.SubcatchAttribute.RUNOFF_RATE, 0, 10)

    assert len(test_array) == 10
    assert np.allclose(test_array, ref_array)


def test_getsubcatchattribute(handle):

    ref_array = np.array([0.125,
                          0.125,
                          0.125,
                          0.125,
                          0.125,
                          0.225,
                          0.225,
                          0.225])

    test_array = output.get_subcatch_attribute(handle, 1, shared_enum.SubcatchAttribute.INFIL_LOSS)

    assert len(test_array) == 8
    assert np.allclose(test_array, ref_array)


def test_getsubcatchresult(handle):

    ref_array = np.array([0.5,
                          0.0,
                          0.0,
                          0.125,
                          1.2438242,
                          0.0,
                          0.0,
                          0.0,
                          33.481991,
                          6.6963983])

    test_array = output.get_subcatch_result(handle, 1, 1)

    assert len(test_array) == 10
    assert np.allclose(test_array, ref_array)


def test_getnoderesult(handle):

    ref_array = np.array([0.296234,
                          995.296204,
                          0.0,
                          1.302650,
                          1.302650,
                          0.0,
                          15.361463,
                          3.072293])

    test_array = output.get_node_result(handle, 2, 2)

    assert len(test_array) == 8
    assert np.allclose(test_array, ref_array)


def test_getlinkresult(handle):

    ref_array = np.array([4.631762,
                          1.0,
                          5.8973422,
                          314.15927,
                          1.0,
                          19.070757,
                          3.8141515])

    test_array = output.get_link_result(handle, 3, 3)

    assert len(test_array) == 7
    assert np.allclose(test_array, ref_array)


def test_getsystemresult(handle):

    ref_array = np.array([70.0,
                          0.1,
                          0.0,
                          0.19042271,
                          14.172027,
                          0.0,
                          0.0,
                          0.0,
                          0.0,
                          14.172027,
                          0.55517411,
                          13.622702,
                          2913.0793,
                          0.0])

    test_array = output.get_system_result(handle, 4, 4)

    assert len(test_array) == 14
    assert np.allclose(test_array, ref_array)

def test_getsystemattribute(handle):

    ref_array = np.array([70.0])

    test_array = output.get_system_attribute(handle, 0, shared_enum.SystemAttribute.AIR_TEMP)

    assert len(test_array) == 1
    assert np.allclose(test_array, ref_array)



#
#  output_metadata.py -
#
#  Created:  February 20, 2020
#  Updated:
#
#  Author:   See AUTHORS
#


from itertools import islice

from aenum import extend_enum

from swmm.toolkit import output, shared_enum


class OutputMetadata:
    '''
    Simple attribute name and unit lookup.
    '''
    _unit_labels_us_ = {
        shared_enum.BaseUnits.RAIN_INT:   "in/hr",
        shared_enum.BaseUnits.SNOW_DEPTH: "in",
        shared_enum.BaseUnits.EVAP_RATE:  "in/day",
        shared_enum.BaseUnits.INFIL_RATE: "in/hr",
        shared_enum.BaseUnits.ELEV:       "ft",
        shared_enum.BaseUnits.PERCENT:    "%",
        shared_enum.BaseUnits.HEAD:       "ft",
        shared_enum.BaseUnits.VOLUME:     "cu ft",
        shared_enum.BaseUnits.VELOCITY:   "ft/sec",
        shared_enum.BaseUnits.TEMP:       "deg F",
        shared_enum.BaseUnits.UNITLESS:   "unitless",
        shared_enum.BaseUnits.NONE:       "",

        shared_enum.FlowUnits.CFS:        "cu ft/sec",
        shared_enum.FlowUnits.GPM:        "gal/min",
        shared_enum.FlowUnits.MGD:        "M gal/day",
    }

    _unit_labels_si_ = {
        shared_enum.BaseUnits.RAIN_INT:   "mm/hr",
        shared_enum.BaseUnits.SNOW_DEPTH: "mm",
        shared_enum.BaseUnits.EVAP_RATE:  "mm/day",
        shared_enum.BaseUnits.INFIL_RATE: "mm/hr",
        shared_enum.BaseUnits.ELEV:       "m",
        shared_enum.BaseUnits.PERCENT:    "%",
        shared_enum.BaseUnits.HEAD:       "m",
        shared_enum.BaseUnits.VOLUME:     "cu m",
        shared_enum.BaseUnits.VELOCITY:   "m/sec",
        shared_enum.BaseUnits.TEMP:       "deg C",
        shared_enum.BaseUnits.UNITLESS:   "unitless",
        shared_enum.BaseUnits.NONE:       "",

        shared_enum.FlowUnits.CMS:        "cu m/sec",
        shared_enum.FlowUnits.LPS:        "L/sec",
        shared_enum.FlowUnits.MLD:        "M L/day",
    }

    _unit_labels_quality_ = {
        shared_enum.ConcUnits.MG:         "mg/L",
        shared_enum.ConcUnits.UG:         "ug/L",
        shared_enum.ConcUnits.COUNT:      "Count/L",
        shared_enum.ConcUnits.NONE:       ""
    }


    def _build_pollut_metadata(self, output_handle):
        '''
        Builds metadata for pollutant attributes at runtime.
        '''
        # Get number of pollutants
        n = output.get_proj_size(output_handle)[shared_enum.ElementType.POLLUT]

        if n > 0:

            pollut_name = []
            pollut_units = []

            # Get pollutant names
            for i in range(0, n):
                pollut_name.append(output.get_elem_name(
                    output_handle, shared_enum.ElementType.POLLUT, i))
            # Get pollutant units
            for u in output.get_units(output_handle)[2:]:
                pollut_units.append(shared_enum.ConcUnits(u))

            # Create dictionary keys
            for i in range(1, n):
                symbolic_name = 'POLLUT_CONC_' + str(i)
                extend_enum(shared_enum.SubcatchAttribute, symbolic_name, 8 + i)
                extend_enum(shared_enum.NodeAttribute, symbolic_name, 6 + i)
                extend_enum(shared_enum.LinkAttribute, symbolic_name, 5 + i)

            # Update metadata dictionary with pollutant metadata
            for i, attr in enumerate(islice(shared_enum.SubcatchAttribute, 8, None)):
                self._metadata[attr] = (pollut_name[i], self._unit_labels[pollut_units[i]])

            for i, attr in enumerate(islice(shared_enum.NodeAttribute, 6, None)):
                self._metadata[attr] = (pollut_name[i], self._unit_labels[pollut_units[i]])

            for i, attr in enumerate(islice(shared_enum.LinkAttribute, 5, None)):
                self._metadata[attr] = (pollut_name[i], self._unit_labels[pollut_units[i]])


    def __init__(self, output_handle):
        # Get units from binary output file
        self.units = output.get_units(output_handle)

        # Determine prevailing unit system
        self._unit_system = shared_enum.UnitSystem(self.units[0])
        if self._unit_system == shared_enum.UnitSystem.US:
            self._unit_labels = type(self)._unit_labels_us_
        else:
            self._unit_labels = type(self)._unit_labels_si_

        self._unit_labels.update(type(self)._unit_labels_quality_)

        # Set user flow units
        self._flow = shared_enum.FlowUnits(self.units[1])

        self._metadata = {
            shared_enum.SubcatchAttribute.RAINFALL:
                ("Rainfall", self._unit_labels[shared_enum.BaseUnits.RAIN_INT]),
            shared_enum.SubcatchAttribute.SNOW_DEPTH:
                ("Snow Depth", self._unit_labels[shared_enum.BaseUnits.SNOW_DEPTH]),
            shared_enum.SubcatchAttribute.EVAP_LOSS:
                ("Evaporation Loss", self._unit_labels[shared_enum.BaseUnits.EVAP_RATE]),
            shared_enum.SubcatchAttribute.INFIL_LOSS:
                ("Infiltration Loss", self._unit_labels[shared_enum.BaseUnits.INFIL_RATE]),
            shared_enum.SubcatchAttribute.RUNOFF_RATE:
                ("Runoff Rate", self._unit_labels[self._flow]),
            shared_enum.SubcatchAttribute.GW_OUTFLOW_RATE:
                ("Groundwater Flow Rate", self._unit_labels[self._flow]),
            shared_enum.SubcatchAttribute.GW_TABLE_ELEV:
                ("Groundwater Elevation", self._unit_labels[shared_enum.BaseUnits.ELEV]),
            shared_enum.SubcatchAttribute.SOIL_MOISTURE:
                ("Soil Moisture", self._unit_labels[shared_enum.BaseUnits.PERCENT]),
            shared_enum.SubcatchAttribute.POLLUT_CONC_0:
                ("Pollutant Concentration", self._unit_labels[shared_enum.BaseUnits.NONE]),

            shared_enum.NodeAttribute.INVERT_DEPTH:
                ("Invert Depth", self._unit_labels[shared_enum.BaseUnits.ELEV]),
            shared_enum.NodeAttribute.HYDRAULIC_HEAD:
                ("Hydraulic Head", self._unit_labels[shared_enum.BaseUnits.HEAD]),
            shared_enum.NodeAttribute.PONDED_VOLUME:
                ("Ponded Volume", self._unit_labels[shared_enum.BaseUnits.VOLUME]),
            shared_enum.NodeAttribute.LATERAL_INFLOW:
                ("Lateral Inflow", self._unit_labels[self._flow]),
            shared_enum.NodeAttribute.TOTAL_INFLOW:
                ("Total Inflow", self._unit_labels[self._flow]),
            shared_enum.NodeAttribute.FLOODING_LOSSES:
                ("Flooding Loss", self._unit_labels[self._flow]),
            shared_enum.NodeAttribute.POLLUT_CONC_0:
                ("Pollutant Concentration", self._unit_labels[shared_enum.BaseUnits.NONE]),

            shared_enum.LinkAttribute.FLOW_RATE:
                ("Flow Rate", self._unit_labels[self._flow]),
            shared_enum.LinkAttribute.FLOW_DEPTH:
                ("Flow Depth", self._unit_labels[shared_enum.BaseUnits.ELEV]),
            shared_enum.LinkAttribute.FLOW_VELOCITY:
                ("Flow Velocity", self._unit_labels[shared_enum.BaseUnits.VELOCITY]),
            shared_enum.LinkAttribute.FLOW_VOLUME:
                ("Flow Volume", self._unit_labels[shared_enum.BaseUnits.VOLUME]),
            shared_enum.LinkAttribute.CAPACITY:
                ("Capacity", self._unit_labels[shared_enum.BaseUnits.PERCENT]),
            shared_enum.LinkAttribute.POLLUT_CONC_0:
                ("Pollutant Concentration", self._unit_labels[shared_enum.BaseUnits.NONE]),

            shared_enum.SystemAttribute.AIR_TEMP:
                ("Temperature", self._unit_labels[shared_enum.BaseUnits.TEMP]),
            shared_enum.SystemAttribute.RAINFALL:
                ("Rainfall", self._unit_labels[shared_enum.BaseUnits.RAIN_INT]),
            shared_enum.SystemAttribute.SNOW_DEPTH:
                ("Snow Depth", self._unit_labels[shared_enum.BaseUnits.SNOW_DEPTH]),
            shared_enum.SystemAttribute.EVAP_INFIL_LOSS:
                ("Evap and Infil Losses", self._unit_labels[shared_enum.BaseUnits.INFIL_RATE]),
            shared_enum.SystemAttribute.RUNOFF_FLOW:
                ("Runoff Flow Rate", self._unit_labels[self._flow]),
            shared_enum.SystemAttribute.DRY_WEATHER_INFLOW:
                ("Dry Weather Inflow", self._unit_labels[self._flow]),
            shared_enum.SystemAttribute.GW_INFLOW:
                ("Groundwater Inflow", self._unit_labels[self._flow]),
            shared_enum.SystemAttribute.RDII_INFLOW:
                ("RDII Inflow", self._unit_labels[self._flow]),
            shared_enum.SystemAttribute.DIRECT_INFLOW:
                ("Direct Inflow", self._unit_labels[self._flow]),
            shared_enum.SystemAttribute.TOTAL_LATERAL_INFLOW:
                ("Total Lateral Inflow", self._unit_labels[self._flow]),
            shared_enum.SystemAttribute.FLOOD_LOSSES:
                ("Flood Losses", self._unit_labels[self._flow]),
            shared_enum.SystemAttribute.OUTFALL_FLOWS:
                ("Outfall Flow", self._unit_labels[self._flow]),
            shared_enum.SystemAttribute.VOLUME_STORED:
                ("Volume Stored", self._unit_labels[shared_enum.BaseUnits.VOLUME]),
            shared_enum.SystemAttribute.EVAP_RATE:
                ("Evaporation Rate", self._unit_labels[shared_enum.BaseUnits.EVAP_RATE]),
            shared_enum.SystemAttribute.PTNL_EVAP_RATE:
                ("Potential Evaporation Rate", self._unit_labels[shared_enum.BaseUnits.EVAP_RATE])
        }

        self._build_pollut_metadata(output_handle)


    def get_attribute_metadata(self, attribute):
        '''
        Takes an attribute enum and returns the name and units in a tuple.
        '''
        return self._metadata[attribute]


# Units of Measurement
#
# Units                  US Customary                           SI Metric
#    AREA_SUBCATCH          acres                 ac               hectares            ha
#    AREA_STOR              square feet           sq ft            square meters       sq m
#    AREA_POND              square feet           sq ft            square meters       sq m
#    CAP_SUC                inches                in               millimeters         mm
#    CONC                   milligrams/liter      mg/L             milligrams/liter    mg/L
#                           micrograms/liter      ug/L             micrograms/liter    ug/L
#                           counts/liter          Count/L          counts/liter        Count/L
#    INFIL_DECAY            1/hours               1/hrs            1/hours             1/hrs
#    POLLUT_DECAY           1/days                1/days           1/days              1/days
#    DEPRES_STOR            inches                in               millimeters         mm
#    DEPTH                  feet                  ft               meters              m
#    DIAM                   feet                  ft               meters              m
#    DISC_COEFF_ORIF        dimensionless         dimless          dimensionless       dimless
#    DISC_COEFF_WEIR        CFS/foot^n            CFS/ft^n         CMS/meter^n         CMS/m^n
#    ELEV                   feet                  ft               meters              m
#    EVAP_RATE              inches/day            in/day           millimeters/day     mm/day
#    FLOW_RATE              cubic feet/sec        CFS              cubic meter/sec     CMS
#                           gallons/minute        GPM              liter/sec           LPS
#                           million gallons/day   MGD              million liter/day   MLD
#    HEAD                   feet                  ft               meters              m
#    HYD_CONDUCT            inches/hour           in/hr            millimeters/hour    mm/hr
#    INFIL_RATE             inches/hour           in/hr            millimeters/hour    mm/hr
#    LEN                    feet                  ft               meters              m
#    MANN_N                 seconds/meter^1/3     sec/m^1/3        seconds/meter^1/3   sec/m^1/3
#    POLLUT_BUILDUP         mass/length           mass/len         mass/length         mass/len
#                           mass/acre             mass/ac          mass/hectare        mass/ha
#    RAIN_INTENSITY         inches/hour           in/hr            millimeters/hour    mm/hr
#    RAIN_VOLUME            inches                in               millimeters         mm
#    SLOPE_SUBCATCH         percent               percent          percent             percent
#    SLOPE_XSEC             rise/run              rise/run         rise/run            rise/run
#    STREET_CLEAN_INT       days                  days             days                days
#    VOLUME                 cubic feet            cu ft            cubic meters        cu m
#    WIDTH                  feet                  ft               meters              m



#
#  test_stats.py
#
#  Created:    Nov  16, 2020
#  Updated:
#
#  Author:     See AUTHORS
#

import os

import pytest

from swmm.toolkit import solver, shared_enum


DATA_PATH = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'data')
INPUT_FILE_EXAMPLE_1 = os.path.join(DATA_PATH, 'test_Example1.inp')
REPORT_FILE_TEST_1 = os.path.join(DATA_PATH, 'temp_Example.rpt')
OUTPUT_FILE_TEST_1 = os.path.join(DATA_PATH, 'temp_Example.out')


@pytest.fixture()
def before_end(request):
    solver.swmm_open(INPUT_FILE_EXAMPLE_1, REPORT_FILE_TEST_1, OUTPUT_FILE_TEST_1)
    solver.swmm_start(0)
    
    while True:
        time = solver.swmm_step()
        if time == 0:
            break
    
    def close():
        solver.swmm_end()
        solver.swmm_close()

    request.addfinalizer(close)


def test_node_get_stats(before_end):
    while True:
        time = solver.swmm_step()
        if time == 0:
            break

    node_stats = solver.node_get_stats(5)
    assert node_stats.maxDepth  == pytest.approx(1.15, 0.1)


def test_storage_get_stats(before_end):
    node_index = solver.project_get_index(shared_enum.ObjectType.NODE, 'SU1')
    while True:
        time = solver.swmm_step()
        if time == 0:
            break

    stor_stats = solver.storage_get_stats(node_index)
    assert stor_stats.avgVol == pytest.approx(0, 0.1)


def test_outfall_stats(before_end):
    id = '18'

    index = solver.project_get_index(shared_enum.ObjectType.NODE, id)

    stats = solver.outfall_get_stats(index)
    assert stats.totalLoad != None


def test_get_routing_totals(before_end):

    stats = solver.system_get_routing_totals()
    assert stats != None


def test_get_runoff_totals(before_end):

    stats = solver.system_get_runoff_totals()
    assert stats != None



#
#   test_metadata.py
#
#   Created:  Apr 8, 2020
#   Modified: Apr 20, 2020
#
#   Author:    See AUTHORS
#
#   Unit testing for SWMM output metadata using pytest.
#


import os

import pytest

from swmm.toolkit import output, shared_enum, output_metadata


DATA_PATH = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'data')
OUTPUT_FILE_EXAMPLE1 = os.path.join(DATA_PATH, 'test_Example1.out')


@pytest.fixture()
def handle(request):
    _handle = output.init()
    output.open(_handle, OUTPUT_FILE_EXAMPLE1)

    def close():
        output.close(_handle)

    request.addfinalizer(close)
    return _handle


def test_outputmetadata(handle):

    om = output_metadata.OutputMetadata(handle)

    ref = {
        shared_enum.SubcatchAttribute.RAINFALL:           ("Rainfall",                   "in/hr"),
        shared_enum.SubcatchAttribute.SNOW_DEPTH:         ("Snow Depth",                 "in"),
        shared_enum.SubcatchAttribute.EVAP_LOSS:          ("Evaporation Loss",           "in/day"),
        shared_enum.SubcatchAttribute.INFIL_LOSS:         ("Infiltration Loss",          "in/hr"),
        shared_enum.SubcatchAttribute.RUNOFF_RATE:        ("Runoff Rate",                "cu ft/sec"),
        shared_enum.SubcatchAttribute.GW_OUTFLOW_RATE:    ("Groundwater Flow Rate",      "cu ft/sec"),
        shared_enum.SubcatchAttribute.GW_TABLE_ELEV:      ("Groundwater Elevation",      "ft"),
        shared_enum.SubcatchAttribute.SOIL_MOISTURE:      ("Soil Moisture",              "%"),
        shared_enum.SubcatchAttribute.POLLUT_CONC_0:      ("TSS",                        "mg/L"),
        shared_enum.SubcatchAttribute.POLLUT_CONC_1:      ("Lead",                       "ug/L"),

        shared_enum.NodeAttribute.INVERT_DEPTH:           ("Invert Depth",               "ft"),
        shared_enum.NodeAttribute.HYDRAULIC_HEAD:         ("Hydraulic Head",             "ft"),
        shared_enum.NodeAttribute.PONDED_VOLUME:          ("Ponded Volume",              "cu ft"),
        shared_enum.NodeAttribute.LATERAL_INFLOW:         ("Lateral Inflow",             "cu ft/sec"),
        shared_enum.NodeAttribute.TOTAL_INFLOW:           ("Total Inflow",               "cu ft/sec"),
        shared_enum.NodeAttribute.FLOODING_LOSSES:        ("Flooding Loss",              "cu ft/sec"),
        shared_enum.NodeAttribute.POLLUT_CONC_0:          ("TSS",                        "mg/L"),
        shared_enum.NodeAttribute.POLLUT_CONC_1:          ("Lead",                       "ug/L"),

        shared_enum.LinkAttribute.FLOW_RATE:              ("Flow Rate",                  "cu ft/sec"),
        shared_enum.LinkAttribute.FLOW_DEPTH:             ("Flow Depth",                 "ft"),
        shared_enum.LinkAttribute.FLOW_VELOCITY:          ("Flow Velocity",              "ft/sec"),
        shared_enum.LinkAttribute.FLOW_VOLUME:            ("Flow Volume",                "cu ft"),
        shared_enum.LinkAttribute.CAPACITY:               ("Capacity",                   "%"),
        shared_enum.LinkAttribute.POLLUT_CONC_0:          ("TSS",                        "mg/L"),
        shared_enum.LinkAttribute.POLLUT_CONC_1:          ("Lead",                       "ug/L"),

        shared_enum.SystemAttribute.AIR_TEMP:             ("Temperature",                "deg F"),
        shared_enum.SystemAttribute.RAINFALL:             ("Rainfall",                   "in/hr"),
        shared_enum.SystemAttribute.SNOW_DEPTH:           ("Snow Depth",                 "in"),
        shared_enum.SystemAttribute.EVAP_INFIL_LOSS:      ("Evap and Infil Losses",      "in/hr"),
        shared_enum.SystemAttribute.RUNOFF_FLOW:          ("Runoff Flow Rate",           "cu ft/sec"),
        shared_enum.SystemAttribute.DRY_WEATHER_INFLOW:   ("Dry Weather Inflow",         "cu ft/sec"),
        shared_enum.SystemAttribute.GW_INFLOW:            ("Groundwater Inflow",         "cu ft/sec"),
        shared_enum.SystemAttribute.RDII_INFLOW:          ("RDII Inflow",                "cu ft/sec"),
        shared_enum.SystemAttribute.DIRECT_INFLOW:        ("Direct Inflow",              "cu ft/sec"),
        shared_enum.SystemAttribute.TOTAL_LATERAL_INFLOW: ("Total Lateral Inflow",       "cu ft/sec"),
        shared_enum.SystemAttribute.FLOOD_LOSSES:         ("Flood Losses",               "cu ft/sec"),
        shared_enum.SystemAttribute.OUTFALL_FLOWS:        ("Outfall Flow",               "cu ft/sec"),
        shared_enum.SystemAttribute.VOLUME_STORED:        ("Volume Stored",              "cu ft"),
        shared_enum.SystemAttribute.EVAP_RATE:            ("Evaporation Rate",           "in/day"),
        shared_enum.SystemAttribute.PTNL_EVAP_RATE:       ("Potential Evaporation Rate", "in/day")
    }

    for attr in shared_enum.SubcatchAttribute:
        temp = om.get_attribute_metadata(attr)
        assert temp == ref[attr]

    for attr in shared_enum.NodeAttribute:
        temp = om.get_attribute_metadata(attr)
        assert temp == ref[attr]

    for attr in shared_enum.LinkAttribute:
        temp = om.get_attribute_metadata(attr)
        assert temp == ref[attr]

    for attr in shared_enum.SystemAttribute:
        temp = om.get_attribute_metadata(attr)
        assert temp == ref[attr]



# This file was automatically generated by SWIG (http://www.swig.org).
# Version 4.0.2
#
# Do not make changes to this file unless you know what you are doing--modify
# the SWIG interface file instead.

"""
Output Module

"""

from sys import version_info as _swig_python_version_info
if _swig_python_version_info < (2, 7, 0):
    raise RuntimeError("Python 2.7 or later required")

# Import the low-level C/C++ module
if __package__ or "." in __name__:
    from . import _output
else:
    import _output

try:
    import builtins as __builtin__
except ImportError:
    import __builtin__

def _swig_repr(self):
    try:
        strthis = "proxy of " + self.this.__repr__()
    except __builtin__.Exception:
        strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)


def _swig_setattr_nondynamic_instance_variable(set):
    def set_instance_attr(self, name, value):
        if name == "thisown":
            self.this.own(value)
        elif name == "this":
            set(self, name, value)
        elif hasattr(self, name) and isinstance(getattr(type(self), name), property):
            set(self, name, value)
        else:
            raise AttributeError("You cannot add instance attributes to %s" % self)
    return set_instance_attr


def _swig_setattr_nondynamic_class_variable(set):
    def set_class_attr(cls, name, value):
        if hasattr(cls, name) and not isinstance(getattr(cls, name), property):
            set(cls, name, value)
        else:
            raise AttributeError("You cannot add class attributes to %s" % cls)
    return set_class_attr


def _swig_add_metaclass(metaclass):
    """Class decorator for adding a metaclass to a SWIG wrapped class - a slimmed down version of six.add_metaclass"""
    def wrapper(cls):
        return metaclass(cls.__name__, cls.__bases__, cls.__dict__.copy())
    return wrapper


class _SwigNonDynamicMeta(type):
    """Meta class to enforce nondynamic attributes (no new attributes) for a class"""
    __setattr__ = _swig_setattr_nondynamic_class_variable(type.__setattr__)


MAXFILENAME = _output.MAXFILENAME
MAXELENAME = _output.MAXELENAME

def init() -> "SMO_Handle *":
    r"""
    Initialize pointer for output handle.

    Returns
    -------
    p_handle: SMO_Handle *
        A SWMM output handle

    """
    return _output.init()

def close(p_handle: "SMO_Handle") -> "int":
    r"""
    Close binary output file and perform cleanup. 

    Parameters
    ----------
    p_handle: SMO_Handle
        A SWMM output handle

    """
    return _output.close(p_handle)

def open(p_handle: "SMO_Handle", path: "char const *") -> "int":
    r"""
    Open binary output file and reads prologue and epilogue.

    Parameters
    ----------
    p_handle: SMO_Handle
        A SWMM output handle
    path: char const *
        The name of the binary output file to be opened.

    """
    return _output.open(p_handle, path)

def get_version(p_handle: "SMO_Handle") -> "int *":
    r"""
    Get the SWMM version that produced the binary output file. 

    Parameters
    ----------
    p_handle: SMO_Handle
        A SWMM output handle

    Returns
    -------
    version: int *
        The SWMM version number found in the output file prologue.


    """
    return _output.get_version(p_handle)

def get_proj_size(p_handle: "SMO_Handle") -> "int *":
    r"""
    Project element counts.

    Parameters
    ----------
    p_handle: SMO_Handle
        A SWMM output handle

    Returns
    -------
    elementCount:int **
        Array of element count values
    length:int *
        Array size

    """
    return _output.get_proj_size(p_handle)

def get_units(p_handle: "SMO_Handle") -> "int *":
    r"""
    Project unit flags for unit_system, flow, and pollutants.

    Parameters
    ----------
    p_handle: SMO_Handle
        A SWMM output handle

    Returns
    -------
    unitFlag: int **
        Array of unit flag values
    length: int *
        Array length

    """
    return _output.get_units(p_handle)

def get_start_date(p_handle: "SMO_Handle") -> "double *":
    r"""
    Analysis start date.

    Parameters
    ----------
    p_handle: SMO_Handle
        A SWMM output handle

    Returns
    -------
    date: double *
        SWMM simulation start date (encoded)

    """
    return _output.get_start_date(p_handle)

def get_times(p_handle: "SMO_Handle", code: "SMO_time") -> "int *":
    r"""
    Analysis step size and number of periods.

    Parameters
    ----------
    p_handle: SMO_Handle
        A SWMM output handle
    code: SMO_time
        A time parameter code (see :ref: SMO_Time)

    Returns
    -------
    time: int *
        Time value

    """
    return _output.get_times(p_handle, code)

def get_elem_name(p_handle: "SMO_Handle", type: "SMO_elementType", elementIndex: "int") -> "int *":
    r"""
    Given an element index returns the element name.

    Parameters
    ----------
    p_handle: SMO_Handle
        A SWMM output handle
    type: SMO_elementType
        The type of the element being queried
    elementIndex: int
        The index of the element being queried

    Returns
    -------
    name: char **
        Element name array
    length: int *
        Length of array

    """
    return _output.get_elem_name(p_handle, type, elementIndex)

def get_subcatch_series(p_handle: "SMO_Handle", subcatchIndex: "int", attr: "SMO_subcatchAttribute", startPeriod: "int", endPeriod: "int") -> "int *":
    r"""
    Get time series results for particular attribute. Specify series start and 
    length using timeIndex and length respectively.

    Parameters
    ----------
    p_handle: SMO_Handle
        A SWMM output handle
    subcatchIndex: int
        The subcatchment index for time series
    attr: SMO_subcatchAttribute
        Code for the subcatch attribute to report (see :ref: SMO_subcatchAttribute)
    startPeriod: int
        Time period of series start
    endPeriod: int
        Time period of series end

    Returns
    -------
    outValueArray: float **
        Array of time series values
    length: int *
        Length of array

    """
    return _output.get_subcatch_series(p_handle, subcatchIndex, attr, startPeriod, endPeriod)

def get_node_series(p_handle: "SMO_Handle", nodeIndex: "int", attr: "SMO_nodeAttribute", startPeriod: "int", endPeriod: "int") -> "int *":
    r"""
    Get time series results for particular attribute. Specify series start and 
    length using timeIndex and length respectively.

    Parameters
    ----------
    p_handle: SMO_Handle
        A SWMM output handle
    nodeIndex: int
        The node index for time series
    attr: SMO_nodeAttribute
        Code for the node attribute to report (see :ref: SMO_nodeAttribute)
    startPeriod: int
        Time period of series start
    endPeriod: int
        Time period of series end

    Returns
    -------
    outValueArray: float **
        Array of time series values
    length: int *
        Length of array

    """
    return _output.get_node_series(p_handle, nodeIndex, attr, startPeriod, endPeriod)

def get_link_series(p_handle: "SMO_Handle", linkIndex: "int", attr: "SMO_linkAttribute", startPeriod: "int", endPeriod: "int") -> "int *":
    r"""
    Get time series results for particular attribute. Specify series start and 
    length using timeIndex and length respectively.

    Parameters
    ----------
    p_handle: SMO_Handle
        A SWMM output handle
    linkIndex: int
        The link index for time series
    attr: SMO_linkAttribute
        Code for the link attribute to report (see :ref: SMO_linkAttribute)
    startPeriod: int
        Time period of series start
    endPeriod: int
        Time period of series end

    Returns
    -------
    outValueArray: float **
        Array of time series values
    length: int *
        Length of array

    """
    return _output.get_link_series(p_handle, linkIndex, attr, startPeriod, endPeriod)

def get_system_series(p_handle: "SMO_Handle", attr: "SMO_systemAttribute", startPeriod: "int", endPeriod: "int") -> "int *":
    r"""
    Get time series results for particular attribute. Specify series start and 
    length using timeIndex and length respectively.

    Parameters
    ----------
    p_handle: SMO_Handle
        A SWMM output handle
    attr: SMO_systemAttribute
        Code for the system attribute to report (see :ref: SMO_systemAttribute)
    startPeriod: int
        Time period of series start
    endPeriod: int
        Time period of series end

    Returns
    -------
    outValueArray: float **
        Array of time series values
    length: int *
        Length of array

    """
    return _output.get_system_series(p_handle, attr, startPeriod, endPeriod)

def get_subcatch_attribute(p_handle: "SMO_Handle", timeIndex: "int", attr: "SMO_subcatchAttribute") -> "int *":
    r"""
    For all subcatchments at given time, get a particular attribute.

    Parameters
    ----------
    p_handle: SMO_Handle
        A SWMM output handle
    timeIndex: int
        The time period index for reporting results
    attr: SMO_subcatchAttribute
        Code for the subcatchment attribute to report (see :ref: SMO_subcatchAttribute)

    Returns
    -------
    outValueArray: float **
        The array of subcatchment attribute values
    length: int *
        Length of array

    """
    return _output.get_subcatch_attribute(p_handle, timeIndex, attr)

def get_node_attribute(p_handle: "SMO_Handle", timeIndex: "int", attr: "SMO_nodeAttribute") -> "int *":
    r"""
    For all nodes at given time, get a particular attribute.

    Parameters
    ----------
    p_handle: SMO_Handle
        A SWMM output handle
    timeIndex: int
        The time period index for reporting results
    attr: SMO_nodeAttribute
        Code for the node attribute to report (see :ref: SMO_nodeAttribute)

    Returns
    -------
    outValueArray: float **
        The array of node attribute values
    length: int *
        Length of array

    """
    return _output.get_node_attribute(p_handle, timeIndex, attr)

def get_link_attribute(p_handle: "SMO_Handle", timeIndex: "int", attr: "SMO_linkAttribute") -> "int *":
    r"""
    For all links at given time, get a particular attribute.

    Parameters
    ----------
    p_handle: SMO_Handle
        A SWMM output handle
    timeIndex: int
        The time period index for reporting results
    attr: SMO_linkAttribute
        Code for the link attribute to report (see :ref: SMO_linkAttribute)

    Returns
    -------
    outValueArray: float **
        The array of link attribute values
    length: int *
        Length of array

    """
    return _output.get_link_attribute(p_handle, timeIndex, attr)

def get_system_attribute(p_handle: "SMO_Handle", timeIndex: "int", attr: "SMO_systemAttribute") -> "int *":
    r"""
    For the system at given time, get a particular attribute.

    Parameters
    ----------
    p_handle: SMO_Handle
        A SWMM output handle
    timeIndex: int
        The time period index for reporting results
    attr: SMO_systemAttribute
        Code for the system attribute to report (see :ref: SMO_systemAttribute)

    Returns
    -------
    outValueArray: float **
        The array of system attribute values
    length: int *
        Length of array

    """
    return _output.get_system_attribute(p_handle, timeIndex, attr)

def get_subcatch_result(p_handle: "SMO_Handle", timeIndex: "int", subcatchIndex: "int") -> "int *":
    r"""
    For a subcatchment at given time, get all attributes.

    Parameters
    ----------
    p_handle: SMO_Handle
        A SWMM output handle
    timeIndex: int
        The time period index for reporting results
    subcatchIndex: int
        The subcatchment index for reporting results

    Returns
    -------
    outValueArray: float **
        The array of subcatchment result values
    length: int *
        Length of array

    """
    return _output.get_subcatch_result(p_handle, timeIndex, subcatchIndex)

def get_node_result(p_handle: "SMO_Handle", timeIndex: "int", nodeIndex: "int") -> "int *":
    r"""
    For a node at given time, get all attributes.

    Parameters
    ----------
    p_handle: SMO_Handle
        A SWMM output handle
    timeIndex: int
        The time period index for reporting results
    nodeIndex: int
        The node index for reporting results

    Returns
    -------
    outValueArray: float **
        The array of node result values
    length: int *
        Length of array

    """
    return _output.get_node_result(p_handle, timeIndex, nodeIndex)

def get_link_result(p_handle: "SMO_Handle", timeIndex: "int", linkIndex: "int") -> "int *":
    r"""
    For a link at given time, get all attributes.

    Parameters
    ----------
    p_handle: SMO_Handle
        A SWMM output handle
    timeIndex: int
        The time period index for reporting results
    linkIndex: int
        The link index for reporting results

    Returns
    -------
    outValueArray: float **
        The array of link result values
    length: int *
        Length of array

    """
    return _output.get_link_result(p_handle, timeIndex, linkIndex)

def get_system_result(p_handle: "SMO_Handle", timeIndex: "int", dummyIndex: "int") -> "int *":
    r"""
    For the system at given time, get all attributes.

    Parameters
    ----------
    p_handle: SMO_Handle
        A SWMM output handle
    timeIndex: int
        The time period index for reporting results
    dummyIndex: int
        No-op for API consistency

    Returns
    -------
    outValueArray: float **
        The array of system result values
    length: int *
        Length of array

    """
    return _output.get_system_result(p_handle, timeIndex, dummyIndex)





# This file was automatically generated by SWIG (http://www.swig.org).
# Version 4.0.2
#
# Do not make changes to this file unless you know what you are doing--modify
# the SWIG interface file instead.

"""
Solver Module

"""

from sys import version_info as _swig_python_version_info
if _swig_python_version_info < (2, 7, 0):
    raise RuntimeError("Python 2.7 or later required")

# Import the low-level C/C++ module
if __package__ or "." in __name__:
    from . import _solver
else:
    import _solver

try:
    import builtins as __builtin__
except ImportError:
    import __builtin__

def _swig_repr(self):
    try:
        strthis = "proxy of " + self.this.__repr__()
    except __builtin__.Exception:
        strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)


def _swig_setattr_nondynamic_instance_variable(set):
    def set_instance_attr(self, name, value):
        if name == "thisown":
            self.this.own(value)
        elif name == "this":
            set(self, name, value)
        elif hasattr(self, name) and isinstance(getattr(type(self), name), property):
            set(self, name, value)
        else:
            raise AttributeError("You cannot add instance attributes to %s" % self)
    return set_instance_attr


def _swig_setattr_nondynamic_class_variable(set):
    def set_class_attr(cls, name, value):
        if hasattr(cls, name) and not isinstance(getattr(cls, name), property):
            set(cls, name, value)
        else:
            raise AttributeError("You cannot add class attributes to %s" % cls)
    return set_class_attr


def _swig_add_metaclass(metaclass):
    """Class decorator for adding a metaclass to a SWIG wrapped class - a slimmed down version of six.add_metaclass"""
    def wrapper(cls):
        return metaclass(cls.__name__, cls.__bases__, cls.__dict__.copy())
    return wrapper


class _SwigNonDynamicMeta(type):
    """Meta class to enforce nondynamic attributes (no new attributes) for a class"""
    __setattr__ = _swig_setattr_nondynamic_class_variable(type.__setattr__)


class NodeStats(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr
    avgDepth = property(_solver.NodeStats_avgDepth_get)
    maxDepth = property(_solver.NodeStats_maxDepth_get)
    maxDepthDate = property(_solver.NodeStats_maxDepthDate_get)
    maxRptDepth = property(_solver.NodeStats_maxRptDepth_get)
    volFlooded = property(_solver.NodeStats_volFlooded_get)
    timeFlooded = property(_solver.NodeStats_timeFlooded_get)
    timeSurcharged = property(_solver.NodeStats_timeSurcharged_get)
    timeCourantCritical = property(_solver.NodeStats_timeCourantCritical_get)
    totLatFlow = property(_solver.NodeStats_totLatFlow_get)
    maxLatFlow = property(_solver.NodeStats_maxLatFlow_get)
    maxInflow = property(_solver.NodeStats_maxInflow_get)
    maxOverflow = property(_solver.NodeStats_maxOverflow_get)
    maxPondedVol = property(_solver.NodeStats_maxPondedVol_get)
    nonConvergedCount = property(_solver.NodeStats_nonConvergedCount_get)
    maxInflowDate = property(_solver.NodeStats_maxInflowDate_get)
    maxOverflowDate = property(_solver.NodeStats_maxOverflowDate_get)

    def __init__(self):
        r"""
        Node Statistics

        Attributes
        ----------
        avgDepth: double
            average node depth (level)
        maxDepth: double
            max node depth (level) (from routing step)
        maxDepthDate: DateTime
            date of maximum depth
        maxRptDepth: double
            max node depth (level) (from reporting step)
        volFlooded: double
            total volume flooded (volume)
        timeFlooded: double
            total time flooded
        timeSurcharged: double
            total time surcharged
        timeCourantCritical: double
            total time courant critical
        totLatFlow: double
            total lateral inflow (volume)
        maxLatFlow: double
            maximum lateral inflow (flowrate)
        maxInflow: double
            maximum total inflow (flowrate)
        maxOverflow: double
            maximum flooding (flowrate)
        maxPondedVol: double
            maximum ponded volume (volume)
        maxInflowDate: DateTime
            date of maximum inflow
        maxOverflowDate: DateTime
            date of maximum overflow

        """
        _solver.NodeStats_swiginit(self, _solver.new_NodeStats())
    __swig_destroy__ = _solver.delete_NodeStats

# Register NodeStats in _solver:
_solver.NodeStats_swigregister(NodeStats)

class StorageStats(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr
    initVol = property(_solver.StorageStats_initVol_get)
    avgVol = property(_solver.StorageStats_avgVol_get)
    maxVol = property(_solver.StorageStats_maxVol_get)
    maxFlow = property(_solver.StorageStats_maxFlow_get)
    evapLosses = property(_solver.StorageStats_evapLosses_get)
    exfilLosses = property(_solver.StorageStats_exfilLosses_get)
    maxVolDate = property(_solver.StorageStats_maxVolDate_get)

    def __init__(self):
        r"""
        Storage Statistics

        Attributes
        ----------
        initVol: double
            initial volume (volume)
        avgVol: double
            average volume (volume) (from routing step)
        maxVol: double
            maximum volume (volume) (from routing step)
        maxFlow: double
            maximum total inflow (flowrate) (from routing step)
        evapLosses: double
            evaporation losses (volume)
        exfilLosses: double
            exfiltration losses (volume)
        maxVolDate: DateTime
            date of maximum volume

        """
        _solver.StorageStats_swiginit(self, _solver.new_StorageStats())
    __swig_destroy__ = _solver.delete_StorageStats

# Register StorageStats in _solver:
_solver.StorageStats_swigregister(StorageStats)

class OutfallStats(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def __init__(self, num_pollut: "int"):
        r"""
        Outfall Statistics

        Attributes
        ----------
        avgFlow: double
            average flow (flowrate)
        maxFlow: double
            maximum flow (flowrate) (from routing step)
        totalLoad: double *
            total pollutant load (mass)
        totalPeriods: double
            total simulation steps (from routing step)

        """
        _solver.OutfallStats_swiginit(self, _solver.new_OutfallStats(num_pollut))
    __swig_destroy__ = _solver.delete_OutfallStats

    def get_totalLoad(self, index: "int") -> "double":
        return _solver.OutfallStats_get_totalLoad(self, index)
    avgFlow = property(_solver.OutfallStats_avgFlow_get)
    maxFlow = property(_solver.OutfallStats_maxFlow_get)
    totalLoad = property(_solver.OutfallStats_totalLoad_get)
    totalPeriods = property(_solver.OutfallStats_totalPeriods_get)

# Register OutfallStats in _solver:
_solver.OutfallStats_swigregister(OutfallStats)

class LinkStats(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr
    maxFlow = property(_solver.LinkStats_maxFlow_get)
    maxFlowDate = property(_solver.LinkStats_maxFlowDate_get)
    maxVeloc = property(_solver.LinkStats_maxVeloc_get)
    maxDepth = property(_solver.LinkStats_maxDepth_get)
    maxStreetFilled = property(_solver.LinkStats_maxStreetFilled_get)
    timeNormalFlow = property(_solver.LinkStats_timeNormalFlow_get)
    timeInletControl = property(_solver.LinkStats_timeInletControl_get)
    timeSurcharged = property(_solver.LinkStats_timeSurcharged_get)
    timeFullUpstream = property(_solver.LinkStats_timeFullUpstream_get)
    timeFullDnstream = property(_solver.LinkStats_timeFullDnstream_get)
    timeFullFlow = property(_solver.LinkStats_timeFullFlow_get)
    timeCapacityLimited = property(_solver.LinkStats_timeCapacityLimited_get)
    timeInFlowClass = property(_solver.LinkStats_timeInFlowClass_get)
    timeCourantCritical = property(_solver.LinkStats_timeCourantCritical_get)
    flowTurns = property(_solver.LinkStats_flowTurns_get)
    flowTurnSign = property(_solver.LinkStats_flowTurnSign_get)

    def __init__(self):
        r"""
        Link Statistics

        .. rubric:: 'Flow Classes'

        ===========   ===========================
        Flow Class    Description                
        ===========   ===========================
        DRY           dry conduit
        UP_DRY        upstream end is dry
        DN_DRY        downstream end is dry
        SUBCRITICAL   sub-critical flow
        SUPCRITICAL   super-critical flow
        UP_CRITICAL   free-fall at upstream end
        DN_CRITICAL   free-fall at downstream end
        ===========   ===========================

        Attributes
        ----------
        maxFlow: double
            maximum flow (flowrate) (from routing step)
        maxFlowDate: double
            date of maximum flowrate
        maxVeloc: double
            maximum velocity (from routing step)
        maxDepth: double
            maximum depth (level)
        timeNormalFlow: double
            time in normal flow
        timeInletControl: double
            time under inlet control
        timeSurcharged: double
            time surcharged
        timeFullUpstream: double
            time full upstream
        timeFullDnstream: double
            time full downstream
        timeFullFlow: double
            time full flow
        timeCapacityLimited: double
            time capacity limited
        timeInFlowClass: double[7]
            time in flow class (See: 'Flow Classes')
        timeCourantCritical: double
            time courant critical
        flowTurns: double
            number of flow turns
        flowTurnSign: double
            number of flow turns sign

        """
        _solver.LinkStats_swiginit(self, _solver.new_LinkStats())
    __swig_destroy__ = _solver.delete_LinkStats

# Register LinkStats in _solver:
_solver.LinkStats_swigregister(LinkStats)

class PumpStats(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr
    utilized = property(_solver.PumpStats_utilized_get)
    minFlow = property(_solver.PumpStats_minFlow_get)
    avgFlow = property(_solver.PumpStats_avgFlow_get)
    maxFlow = property(_solver.PumpStats_maxFlow_get)
    volume = property(_solver.PumpStats_volume_get)
    energy = property(_solver.PumpStats_energy_get)
    offCurveLow = property(_solver.PumpStats_offCurveLow_get)
    offCurveHigh = property(_solver.PumpStats_offCurveHigh_get)
    startUps = property(_solver.PumpStats_startUps_get)
    totalPeriods = property(_solver.PumpStats_totalPeriods_get)

    def __init__(self):
        r"""
        Pump Statistics

        Attributes
        ----------
        utilized: double
            time utilized
        minFlow: double
            minimum flowrate
        avgFlow: double
            average flowrate
        maxFlow: double
            maximum flowrate
        volume: double
            total pumping volume (volume)
        energy: double
            total energy demand
        offCurveLow: double
            hysteresis low (off depth wrt curve)
        offCurveHigh: double
            hysteresis high (on depth wrt curve)
        startUps: int
            number of start ups
        totalPeriods: int
            total simulation steps (from routing step)

        """
        _solver.PumpStats_swiginit(self, _solver.new_PumpStats())
    __swig_destroy__ = _solver.delete_PumpStats

# Register PumpStats in _solver:
_solver.PumpStats_swigregister(PumpStats)

class SubcatchStats(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr
    precip = property(_solver.SubcatchStats_precip_get)
    runon = property(_solver.SubcatchStats_runon_get)
    evap = property(_solver.SubcatchStats_evap_get)
    infil = property(_solver.SubcatchStats_infil_get)
    runoff = property(_solver.SubcatchStats_runoff_get)
    maxFlow = property(_solver.SubcatchStats_maxFlow_get)
    impervRunoff = property(_solver.SubcatchStats_impervRunoff_get)
    pervRunoff = property(_solver.SubcatchStats_pervRunoff_get)

    def __init__(self):
        r"""
        Subcatchment Statistics

        Attributes
        ----------
        precip: double
            total precipication (length)
        runon: double
            total runon (volume)
        evap: double
            total evaporation (volume)
        infil: double
            total infiltration (volume)
        runoff: double
            total runoff (volume)
        maxFlow: double
            maximum runoff rate (flowrate)

        """
        _solver.SubcatchStats_swiginit(self, _solver.new_SubcatchStats())
    __swig_destroy__ = _solver.delete_SubcatchStats

# Register SubcatchStats in _solver:
_solver.SubcatchStats_swigregister(SubcatchStats)

class RoutingTotals(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr
    dwInflow = property(_solver.RoutingTotals_dwInflow_get)
    wwInflow = property(_solver.RoutingTotals_wwInflow_get)
    gwInflow = property(_solver.RoutingTotals_gwInflow_get)
    iiInflow = property(_solver.RoutingTotals_iiInflow_get)
    exInflow = property(_solver.RoutingTotals_exInflow_get)
    flooding = property(_solver.RoutingTotals_flooding_get)
    outflow = property(_solver.RoutingTotals_outflow_get)
    evapLoss = property(_solver.RoutingTotals_evapLoss_get)
    seepLoss = property(_solver.RoutingTotals_seepLoss_get)
    reacted = property(_solver.RoutingTotals_reacted_get)
    initStorage = property(_solver.RoutingTotals_initStorage_get)
    finalStorage = property(_solver.RoutingTotals_finalStorage_get)
    pctError = property(_solver.RoutingTotals_pctError_get)

    def __init__(self):
        r"""
        System Flow Routing Totals

        Attributes
        ----------
        dwInflow: double
            dry weather inflow
        wwInflow: double
            wet weather inflow
        gwInflow: double
            groundwater inflow
        iiInflow: double
            RDII inflow
        exInflow: double
            direct inflow
        flooding: double
            internal flooding
        outflow: double
            external outflow
        evapLoss: double
            evaporation loss
        seepLoss: double
            seepage loss
        reacted: double
            reaction losses
        initStorage: double
            initial storage volume
        finalStorage: double
            final storage volume
        pctError: double
            continuity error

        """
        _solver.RoutingTotals_swiginit(self, _solver.new_RoutingTotals())
    __swig_destroy__ = _solver.delete_RoutingTotals

# Register RoutingTotals in _solver:
_solver.RoutingTotals_swigregister(RoutingTotals)

class RunoffTotals(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr
    rainfall = property(_solver.RunoffTotals_rainfall_get)
    evap = property(_solver.RunoffTotals_evap_get)
    infil = property(_solver.RunoffTotals_infil_get)
    runoff = property(_solver.RunoffTotals_runoff_get)
    drains = property(_solver.RunoffTotals_drains_get)
    runon = property(_solver.RunoffTotals_runon_get)
    initStorage = property(_solver.RunoffTotals_initStorage_get)
    finalStorage = property(_solver.RunoffTotals_finalStorage_get)
    initSnowCover = property(_solver.RunoffTotals_initSnowCover_get)
    finalSnowCover = property(_solver.RunoffTotals_finalSnowCover_get)
    snowRemoved = property(_solver.RunoffTotals_snowRemoved_get)
    pctError = property(_solver.RunoffTotals_pctError_get)

    def __init__(self):
        r"""
        System Runoff Totals

        Attributes
        ----------
        rainfall: double
            rainfall total (depth)
        evap: double
            evaporation loss (volume)
        infil: double
            infiltration loss (volume)
        runoff: double
            runoff volume (volume)
        drains: double
            LID drains (volume)
        runon: double
            runon from outfalls (volume)
        initStorage: double
            inital surface storage (depth)
        finalStorage: double
            final surface storage (depth)
        initSnowCover: double
            initial snow cover (depth)
        finalSnowCover: double
            final snow cover (depth)
        snowRemoved: double
            snow removal (depth)
        pctError: double
            continuity error (%)

        """
        _solver.RunoffTotals_swiginit(self, _solver.new_RunoffTotals())
    __swig_destroy__ = _solver.delete_RunoffTotals

# Register RunoffTotals in _solver:
_solver.RunoffTotals_swigregister(RunoffTotals)

swmm_GAGE = _solver.swmm_GAGE
swmm_SUBCATCH = _solver.swmm_SUBCATCH
swmm_NODE = _solver.swmm_NODE
swmm_LINK = _solver.swmm_LINK
swmm_SYSTEM = _solver.swmm_SYSTEM
swmm_JUNCTION = _solver.swmm_JUNCTION
swmm_OUTFALL = _solver.swmm_OUTFALL
swmm_STORAGE = _solver.swmm_STORAGE
swmm_DIVIDER = _solver.swmm_DIVIDER
swmm_CONDUIT = _solver.swmm_CONDUIT
swmm_PUMP = _solver.swmm_PUMP
swmm_ORIFICE = _solver.swmm_ORIFICE
swmm_WEIR = _solver.swmm_WEIR
swmm_OUTLET = _solver.swmm_OUTLET
swmm_GAGE_RAINFALL = _solver.swmm_GAGE_RAINFALL
swmm_SUBCATCH_AREA = _solver.swmm_SUBCATCH_AREA
swmm_SUBCATCH_RAINGAGE = _solver.swmm_SUBCATCH_RAINGAGE
swmm_SUBCATCH_RAINFALL = _solver.swmm_SUBCATCH_RAINFALL
swmm_SUBCATCH_EVAP = _solver.swmm_SUBCATCH_EVAP
swmm_SUBCATCH_INFIL = _solver.swmm_SUBCATCH_INFIL
swmm_SUBCATCH_RUNOFF = _solver.swmm_SUBCATCH_RUNOFF
swmm_SUBCATCH_RPTFLAG = _solver.swmm_SUBCATCH_RPTFLAG
swmm_NODE_TYPE = _solver.swmm_NODE_TYPE
swmm_NODE_ELEV = _solver.swmm_NODE_ELEV
swmm_NODE_MAXDEPTH = _solver.swmm_NODE_MAXDEPTH
swmm_NODE_DEPTH = _solver.swmm_NODE_DEPTH
swmm_NODE_HEAD = _solver.swmm_NODE_HEAD
swmm_NODE_VOLUME = _solver.swmm_NODE_VOLUME
swmm_NODE_LATFLOW = _solver.swmm_NODE_LATFLOW
swmm_NODE_INFLOW = _solver.swmm_NODE_INFLOW
swmm_NODE_OVERFLOW = _solver.swmm_NODE_OVERFLOW
swmm_NODE_RPTFLAG = _solver.swmm_NODE_RPTFLAG
swmm_LINK_TYPE = _solver.swmm_LINK_TYPE
swmm_LINK_NODE1 = _solver.swmm_LINK_NODE1
swmm_LINK_NODE2 = _solver.swmm_LINK_NODE2
swmm_LINK_LENGTH = _solver.swmm_LINK_LENGTH
swmm_LINK_SLOPE = _solver.swmm_LINK_SLOPE
swmm_LINK_FULLDEPTH = _solver.swmm_LINK_FULLDEPTH
swmm_LINK_FULLFLOW = _solver.swmm_LINK_FULLFLOW
swmm_LINK_SETTING = _solver.swmm_LINK_SETTING
swmm_LINK_TIMEOPEN = _solver.swmm_LINK_TIMEOPEN
swmm_LINK_TIMECLOSED = _solver.swmm_LINK_TIMECLOSED
swmm_LINK_FLOW = _solver.swmm_LINK_FLOW
swmm_LINK_DEPTH = _solver.swmm_LINK_DEPTH
swmm_LINK_VELOCITY = _solver.swmm_LINK_VELOCITY
swmm_LINK_TOPWIDTH = _solver.swmm_LINK_TOPWIDTH
swmm_LINK_RPTFLAG = _solver.swmm_LINK_RPTFLAG
swmm_STARTDATE = _solver.swmm_STARTDATE
swmm_CURRENTDATE = _solver.swmm_CURRENTDATE
swmm_ELAPSEDTIME = _solver.swmm_ELAPSEDTIME
swmm_ROUTESTEP = _solver.swmm_ROUTESTEP
swmm_MAXROUTESTEP = _solver.swmm_MAXROUTESTEP
swmm_REPORTSTEP = _solver.swmm_REPORTSTEP
swmm_TOTALSTEPS = _solver.swmm_TOTALSTEPS
swmm_NOREPORT = _solver.swmm_NOREPORT
swmm_FLOWUNITS = _solver.swmm_FLOWUNITS
swmm_CFS = _solver.swmm_CFS
swmm_GPM = _solver.swmm_GPM
swmm_MGD = _solver.swmm_MGD
swmm_CMS = _solver.swmm_CMS
swmm_LPS = _solver.swmm_LPS
swmm_MLD = _solver.swmm_MLD

def swmm_run(f1: "char const *", f2: "char const *", f3: "char const *") -> "int":
    r"""
    Opens SWMM input file, reads in network data, runs, and closes

    Parameters
    ----------
    f1: char const *
        pointer to name of input file (must exist)
    f2: char const *
        pointer to name of report file (to be created)
    f3: char const *
        pointer to name of binary output file (to be created)

    """
    return _solver.swmm_run(f1, f2, f3)

def swmm_open(f1: "char const *", f2: "char const *", f3: "char const *") -> "int":
    r"""
    Opens SWMM input file & reads in network data

    Parameters
    ----------
    f1: char const *
        pointer to name of input file (must exist)
    f2: char const *
        pointer to name of report file (to be created)
    f3: char const *
        pointer to name of binary output file (to be created)

    """
    return _solver.swmm_open(f1, f2, f3)

def swmm_start(saveFlag: "int") -> "int":
    r"""
    Start SWMM simulation

    Parameters
    ----------
    saveFlag: int
        TRUE or FALSE to save timeseries to report file

    """
    return _solver.swmm_start(saveFlag)

def swmm_step() -> "double *":
    r"""
    Step SWMM simulation forward

    Returns
    -------
    elapsedTime: double
        elapsed simulation time [milliseconds]

    """
    return _solver.swmm_step()

def swmm_stride(strideStep: "int") -> "double *":
    return _solver.swmm_stride(strideStep)

def swmm_end() -> "int":
    r"""
    End SWMM simulation   

    """
    return _solver.swmm_end()

def swmm_report() -> "int":
    r"""
    Write text report file

    """
    return _solver.swmm_report()

def swmm_close() -> "int":
    r"""
    Frees all memory and files used by SWMM

    """
    return _solver.swmm_close()

def swmm_get_mass_balance() -> "float *, float *, float *":
    r"""
    Get routing errors

    Returns
    -------
    runoffErr: float *
        Runoff routing error
    flowErr: float * 
        Flow routing error
    qualErr: float * 
        Quality routing error

    """
    return _solver.swmm_get_mass_balance()

def swmm_get_version() -> "int":
    r"""
    Get Legacy SWMM version number

    Returns
    -------
    version: int
        Version number

    """
    return _solver.swmm_get_version()

def swmm_version_info() -> "char const *":
    return _solver.swmm_version_info()

def swmm_build_id() -> "char const *":
    return _solver.swmm_build_id()

def project_get_index(type: "SM_ObjectType", id: "char *") -> "int *":
    r"""
    Finds the index of an object given its ID.

    Parameters
    ----------
    type: SM_ObjectType
        type An object type (see :ref: SM_ObjectType)
    id: char *
        id The object ID

    Returns
    -------    
    index: int *
        The objects index

    """
    return _solver.project_get_index(type, id)

def simulation_get_unit(type: "SM_Units") -> "int *":
    r"""
    Gets Simulation Unit

    Parameters
    ----------
    type: SM_Units
        Option code (see :ref: SM_Units)

    Returns
    -------
    value: int *
        Option value

    """
    return _solver.simulation_get_unit(type)

def simulation_get_setting(type: "SM_SimOption") -> "int *":
    r"""
    Gets Simulation Analysis Setting

    Parameters
    ----------
    type: SM_SimOption
        Option code (see :ref: SM_SimOption)

    Returns
    -------
    value: int *
        Option value

    """
    return _solver.simulation_get_setting(type)

def simulation_get_parameter(type: "SM_SimSetting") -> "double *":
    r"""
    Gets Simulation Analysis Setting

    Parameters
    ----------
    type: SM_SimSetting
        Option code (see :ref: SM_SimSetting)

    Returns
    -------
    value: double *
        Option value

    """
    return _solver.simulation_get_parameter(type)

def project_get_count(type: "SM_ObjectType") -> "int *":
    r"""
    Gets Object Count

    Parameters
    ----------
    type: SM_ObjectType
        Option code (see :ref: SM_ObjectType)

    Returns
    -------
    count: int *
        Option value

    """
    return _solver.project_get_count(type)

def project_get_id(type: "SM_ObjectType", index: "int") -> "char **":
    r"""
    Gets Object ID

    Parameters
    ----------
    type: SM_ObjectType
        type Option code (see :ref: SM_ObjectType)
    index: int
        index of the Object

    Returns
    -------
    id: char **
        The string ID of object.

    """
    return _solver.project_get_id(type, index)

def node_get_type(index: "int") -> "SM_NodeType *":
    r"""
    Get the type of node with specified index.

    Parameters
    ----------
    index: int
        The index of a node

    Returns
    -------
    Ntype: SM_NodeType *
        The type code for the node (:ref: SM_NodeType). 
        id must be pre-allocated by the caller.

    """
    return _solver.node_get_type(index)

def link_get_type(index: "int") -> "SM_LinkType *":
    r"""
    Get the type of link with specified index.

    Parameters
    ----------
    index: int
        The index of a link

    Returns
    -------
    Ltype: SM_LinkType *
        The type code for the link (:ref: SM_LinkType).

    """
    return _solver.link_get_type(index)

def link_get_connections(index: "int") -> "int *, int *":
    r"""
    Get the link Connection Node Indeces. If the conduit has a negative slope, 
    the dynamic wave solver will automatically reverse the nodes. 

    Parameters
    ----------
    index: int
        The index of a link

    Returns
    -------
    Node1: int *
        The upstream node index.
    Node2: int *
        The downstream node index.

    """
    return _solver.link_get_connections(index)

def link_get_direction(index: "int") -> "signed char *":
    r"""
    Get the link flow direction

    Parameters
    ----------
    index: int
        The index of a link

    Returns
    -------
    value: char *
        The link flow direction.

    """
    return _solver.link_get_direction(index)

def subcatch_get_connection(index: "int") -> "SM_ObjectType *, int *":
    r"""
    Get the Subcatchment connection. Subcatchments can load to a node, another 
    subcatchment, or itself.

    Parameters
    ----------
    index: int
        The index of a Subcatchment

    Returns
    -------
    type: SM_ObjectType
        The type of object loading (See :ref: SM_ObjectType)
    out_index: int *
        The object index

    """
    return _solver.subcatch_get_connection(index)

def lid_usage_get_count(index: "int") -> "int *":
    r"""
    Get the number of lid units on a subcatchment.

    Parameters
    ----------
    index: int
        The index of a subcatchment

    Returns
    -------
    value: int *
        The number of lid units on a subcatchment

    """
    return _solver.lid_usage_get_count(index)

def lid_usage_get_parameter(index: "int", lidIndex: "int", param: "SM_LidUProperty") -> "double *":
    r"""
    Get a property value for a specified lid unit on a specified subcatchment

    Parameters
    ----------
    index: int
        The index of a subcatchment
    lidIndex: int
        The index of specified lid unit
    param: SM_LidUProperty
        The property type code (See :ref: SM_LidUProperty)

    Returns
    -------
    value: double
        The value of the lid unit's property

    """
    return _solver.lid_usage_get_parameter(index, lidIndex, param)

def lid_usage_set_parameter(index: "int", lidIndex: "int", param: "SM_LidUProperty", value: "double") -> "int":
    r"""
    Set a property value for a specified lid unit on a specified subcatchment

    Parameters
    ----------
    index: int
        The index of a subcatchment
    lidIndex: int
        The index of specified lid unit
    param: SM_LidUProperty
        The property type code (See :ref: SM_LidUProperty)
    value: double
        The new value of the lid unit's property

    """
    return _solver.lid_usage_set_parameter(index, lidIndex, param, value)

def lid_usage_get_option(index: "int", lidIndex: "int", param: "SM_LidUOptions") -> "int *":
    r"""
    Get the lid option for a specified lid unit on a specified subcatchment

    Parameters
    ----------
    index: int
        The index of a subcatchment
    lidIndex: int
        The index of specified lid unit
    param: SM_LidUOptions
        The lid option type code (See :ref: SM_LidUOptions)

    Returns
    -------
    value: int *
        The value of the option for the lid unit

    """
    return _solver.lid_usage_get_option(index, lidIndex, param)

def lid_usage_set_option(index: "int", lidIndex: "int", param: "SM_LidUOptions", value: "int") -> "int":
    r"""
    Set the lid option for a specified lid unit on a specified subcatchment

    Parameters
    ----------
    index: int
        The index of a subcatchment
    lidIndex: int
        The index of specified lid unit
    param: SM_LidUOptions
        The lid option type code (See :ref: SM_LidUOptions)
    value: int
        The new value of the option for the lid unit

    """
    return _solver.lid_usage_set_option(index, lidIndex, param, value)

def lid_control_get_overflow(lidControlIndex: "int") -> "int *":
    r"""
    Get the lid control surface immediate overflow condition

    Parameters
    ----------
    lidControlIndex: int
        The index of specified lid control

    Returns
    -------
    condition: int *
        The value of surface immediate overflow condition

    """
    return _solver.lid_control_get_overflow(lidControlIndex)

def lid_control_get_parameter(lidControlIndex: "int", layerIndex: "SM_LidLayer", param: "SM_LidLayerProperty") -> "double *":
    r"""
    Get a property value for specified lid control

    Parameters
    ----------
    lidControlIndex: int
        The index of specified lid control
    layerIndex: SM_LidLayer
        The index of specified lid layer (See :ref: SM_LidLayer)
    param: SM_LidLayerProperty
        The property type code (See :ref: SM_LidLayerProperty)

    Returns
    -------
    value: double 
        The value of lid control's property

    """
    return _solver.lid_control_get_parameter(lidControlIndex, layerIndex, param)

def lid_control_set_parameter(lidControlIndex: "int", layerIndex: "SM_LidLayer", param: "SM_LidLayerProperty", value: "double") -> "int":
    r"""
    Set a property value for specified lid control

    Parameters
    ----------
    lidControlIndex: int
        The index of specified lid control
    layerIndex: SM_LidLayer
        The index of specified lid layer (See :ref: SM_LidLayer)
    param: SM_LidLayerProperty
        The property type code (See :ref: SM_LidLayerProperty)
    value: double
        The new value for the lid control's property

    """
    return _solver.lid_control_set_parameter(lidControlIndex, layerIndex, param, value)

def lid_usage_get_flux_rate(index: "int", lidIndex: "int", layerIndex: "SM_LidLayer") -> "double *":
    r"""
    Get the lid unit water balance simulated value at current time

    Parameters
    ----------
    index: int
        The index of a subcatchment
    lidIndex: int
        The index of specified lid unit
    layerIndex: SM_LidLayer
        The index of specified lid layer (See :ref: SM_LidLayer)

    Returns
    -------
    result: double * 
        The result for the specified lid unit

    """
    return _solver.lid_usage_get_flux_rate(index, lidIndex, layerIndex)

def lid_group_get_result(index: "int", type: "SM_LidResult") -> "double *":
    r"""
    Get the lid group of a specified subcatchment result at current time

    Parameters
    ----------
    index: int
        The index of a subcatchment
    type: SM_LidResult
        The result type code (See :ref: SM_LidResult)

    Returns
    -------
    result: double *
        The result for the specified lid group

    """
    return _solver.lid_group_get_result(index, type)

def lid_usage_get_result(index: "int", lidIndex: "int", type: "SM_LidResult") -> "double *":
    r"""
    Get the lid unit of a specified subcatchment result at current time

    Parameters
    ----------
    index: int
        The index of a subcatchment
    lidIndex: int
        The index of specified lid unit
    type: SM_LidResult
        The result type code (See :ref: SM_LidResult)

    Returns
    -------
    result: double *
        The result for the specified lid unit

    """
    return _solver.lid_usage_get_result(index, lidIndex, type)

def node_get_parameter(index: "int", param: "SM_NodeProperty") -> "double *":
    r"""
    Get a property value for specified node.

    Parameters
    ----------
    index: int
        The index of a node
    param: SM_NodeProperty
        The property type code (See :ref: SM_NodeProperty)

    Returns
    -------
    value: double *
        The value of the node's property

    """
    return _solver.node_get_parameter(index, param)

def node_set_parameter(index: "int", param: "SM_NodeProperty", value: "double") -> "int":
    r"""
    Set a property value for specified node.

    Parameters
    ----------
    index: int
        The index of a node
    param: SM_NodeProperty
        The property type code (See :ref: SM_NodeProperty)
    value: double
        The new value of the node's property

    """
    return _solver.node_set_parameter(index, param, value)

def link_get_parameter(index: "int", param: "SM_LinkProperty") -> "double *":
    r"""
    Get a property value for specified link.

    Parameters
    ----------
    index: int
        The index of a link
    param: SM_LinkProperty
        The property type code (See :ref: SM_LinkProperty)

    Returns
    -------
    value: double *
        The value of the link's property

    """
    return _solver.link_get_parameter(index, param)

def link_set_parameter(index: "int", param: "SM_LinkProperty", value: "double") -> "int":
    r"""
    Set a property value for specified link.

    Parameters
    ----------
    index: int
        The index of a link
    param: SM_LinkProperty
        The property type code (See :ref: SM_LinkProperty)
    value: double
        The new value of the link's property

    """
    return _solver.link_set_parameter(index, param, value)

def inlet_get_parameter(index: "int", param: "SM_InletProperty") -> "double *":
    r"""
    Get a property value for the inlet of a specified link.

    Parameters
    ----------
    index: int
        The index of a link with desired inlets
    param: SM_InletProperty
        The property type code (See :ref: SM_InletProperty)

    Returns
    -------
    value: double *
        The value of the inlet's property

    """
    return _solver.inlet_get_parameter(index, param)

def inlet_set_parameter(index: "int", param: "SM_InletProperty", value: "double") -> "int":
    r"""
    Get a property value for the inlet of a specified link.

    Parameters
    ----------
    index: int
        The index of a link with desired inlets
    param: SM_InletProperty
        The property type code (See :ref: SM_InletProperty)

    Returns
    -------
    value: double *
        The new value of the inlet's property

    """
    return _solver.inlet_set_parameter(index, param, value)

def subcatch_get_parameter(index: "int", param: "SM_SubcProperty") -> "double *":
    r"""
    Get a property value for specified subcatchment.

    Parameters
    ----------
    index: int
        The index of a subcatchment
    param: SM_SubcProperty
        The property type code (See :ref: SM_SubcProperty)

    Returns
    -------
    value: double *
        The value of the subcatchment's property

    """
    return _solver.subcatch_get_parameter(index, param)

def subcatch_set_parameter(index: "int", param: "SM_SubcProperty", value: "double") -> "int":
    r"""
    Set a property value for specified subcatchment.

    Parameters
    ----------
    index: int
        The index of a subcatchment
    param: SM_SubcProperty
        The property type code (See :ref: SM_SubcProperty)
    value: double
        The new value of the subcatchment's property

    """
    return _solver.subcatch_set_parameter(index, param, value)

def simulation_get_datetime(type: "SM_TimePropety") -> "int *, int *, int *, int *, int *, int *":
    r"""
    Get the simulation datetime information

    Parameters
    ----------
    type: SM_TimePropety
        The property type code (See :ref: SM_TimePropety)

    Returns
    -------
    year: int *
        The year
    month: int *
        The month
    day: int *
        The day
    hour: int *
        The hour
    minute: int *
        The minute
    second: int *
        The seconds

    """
    return _solver.simulation_get_datetime(type)

def simulation_set_datetime(type: "SM_TimePropety", year: "int", month: "int", day: "int", hour: "int", minute: "int", second: "int") -> "int":
    r"""
    Set simulation datetime information.

    Parameters
    ----------
    type: SM_TimePropety 
        type The property type code (See :ref: SM_TimePropety)
    year: int
        The year
    month: int
        The month
    day: int
        The day
    hour: int
        The hour
    minute: int
        The minute
    second: int
        The seconds

    """
    return _solver.simulation_set_datetime(type, year, month, day, hour, minute, second)

def simulation_get_current_datetime() -> "int *, int *, int *, int *, int *, int *":
    r"""
    Get the current simulation datetime information.

    Returns
    -------
    year: int *
        The year
    month: int *
        The month
    day: int *
        The day
    hour: int *
        The hour
    minute: int *
        The minute
    second: int *
        The seconds

    """
    return _solver.simulation_get_current_datetime()

def node_get_result(index: "int", type: "SM_NodeResult") -> "double *":
    r"""
    Get a result value for specified node.

    Parameters
    ----------
    index: int
        The index of a node
    type: SM_NodeResult
        The property type code (See :ref: SM_NodeResult)

    Returns
    -------
    result: double *
        The result of the node's property

    """
    return _solver.node_get_result(index, type)

def node_get_pollutant(index: "int", type: "SM_NodePollut") -> "int *":
    r"""
    Gets pollutant values for a specified node.

    Parameters
    ----------
    index: int
        The index of a node
    type: SM_NodePollut
        The property type code (see :ref: SM_NodePollut)

    Returns
    -------
    PollutArray: double *
        result array

    """
    return _solver.node_get_pollutant(index, type)

def node_set_pollutant(index: "int", type: "SM_NodePollut", pollutant_index: "int", pollutant_value: "double") -> "int":
    r"""
    Set node pollutant concentration

    Parameters
    ----------
    index: int
        The node index.
    type: SM_NodePollut
        The property type code (see :ref: SM_NodePollut). Currently Only QUALITY (0) is supported.
    pollutant_index: int
        The index of the pollutant to set.
    pollutant_value:
        The new pollutant concentration to set.


    """
    return _solver.node_set_pollutant(index, type, pollutant_index, pollutant_value)

def link_set_pollutant(index: "int", type: "SM_LinkPollut", pollutant_index: "int", pollutant_value: "double") -> "int":
    r"""
    Set link pollutant concentration

    Parameters
    ----------
    index: int
        The link index.
    type: SM_LinkPollut
        The property type code (see :ref: SM_LinkPollut). Currently Only QUALITY (0) is supported.
    pollutant_index: int
        The index of the pollutant to set.
    pollutant_value:
        The new pollutant concentration to set.


    """
    return _solver.link_set_pollutant(index, type, pollutant_index, pollutant_value)

def link_get_result(index: "int", type: "SM_LinkResult") -> "double *":
    r"""
    Get a result value for specified link.

    Parameters
    ----------
    index: int
        The index of a link
    type: SM_LinkResult
        The property type code (See :ref: SM_LinkResult)

    Returns
    -------
    result: double *
        The result of the link's property

    """
    return _solver.link_get_result(index, type)

def inlet_get_result(index: "int", type: "SM_InletResult") -> "double *":
    r"""
    Get a result value for specified link.

    Parameters
    ----------
    index: int
        The index of a link with desired inlets
    type: SM_InletResult
        The result type code (See :ref: SM_InletResult)

    Returns
    -------
    result: double *
        The value of the inlet's result at the current simulation timestep

    """
    return _solver.inlet_get_result(index, type)

def link_get_pollutant(index: "int", type: "SM_LinkPollut") -> "int *":
    r"""
    Gets pollutant values for a specified link.

    Parameters
    ----------
    index: int
        The index of a link
    type: SM_LinkPollut
        The property type code (see :ref: SM_LinkPollut)

    Returns
    -------
    PollutArray: double *
        result array

    """
    return _solver.link_get_pollutant(index, type)

def subcatch_get_result(index: "int", type: "SM_SubcResult") -> "double *":
    r"""
    Get a result value for specified subcatchment.

    Parameters
    ----------
    index: int
        The index of a subcatchment
    type: SM_SubcResult
        The property type code (See :ref: SM_SubcResult)

    Returns
    -------
    result: double *
        The result of the subcatchment's property

    """
    return _solver.subcatch_get_result(index, type)

def subcatch_get_pollutant(index: "int", type: "SM_SubcPollut") -> "int *":
    r"""
    Gets pollutant values for a specified subcatchment.

    Parameters
    ----------
    index: int
        The index of a subcatchment
    type: SM_SubcPollut
        The property type code (see :ref: SM_SubcPollut)

    Returns
    -------
    PollutArray: double **
        result array

    """
    return _solver.subcatch_get_pollutant(index, type)

def raingage_get_precipitation(index: "int", type: "SM_GagePrecip") -> "double *":
    r"""
    Get precipitation rates for a gage.

    Parameters
    ----------
    index: int
        The index of gage
    type: SM_GagePrecip
        The property type code (see :ref: SM_GagePrecip)

    Returns
    -------
    GageArray: double *
        precipitation rate

    """
    return _solver.raingage_get_precipitation(index, type)

def node_get_stats(index: "int") -> "SM_NodeStats *":
    r"""
    Get a node statistics.

    Parameters
    ----------
    index: int
        The index of a node

    Returns
    -------
    nodeStats: SM_NodeStats * 
        The Node Stats struct (see :ref: SM_NodeStats) pre-allocated by the caller.

    """
    return _solver.node_get_stats(index)

def node_get_total_inflow(index: "int") -> "double *":
    r"""
    Get the cumulative inflow for a node.

    Parameters
    ----------
    index: int
        The index of a node

    Returns
    -------
    value: double *
        The total inflow.

    """
    return _solver.node_get_total_inflow(index)

def storage_get_stats(index: "int") -> "SM_StorageStats *":
    r"""
    Get a storage statistics.

    Parameters
    ----------
    index: int
        The index of a storage node

    Returns
    -------
    storageStats: SM_StorageStats * 
        The storage Stats struct (see :ref: SM_StorageStats) pre-allocated by the caller.

    """
    return _solver.storage_get_stats(index)

def outfall_get_stats(index: "int") -> "SM_OutfallStats *":
    r"""
    Get outfall statistics.

    Parameters
    ----------
    index: int
        The index of a outfall node

    Returns
    -------
    outfallStats: SM_OutfallStats * 
        The outfall Stats struct (see :ref: SM_OutfallStats).
        pre-allocated by the caller. Caller is also responsible for freeing the
        SM_OutfallStats structure using swmm_freeOutfallStats(). This frees any
        pollutants array.

    """
    return _solver.outfall_get_stats(index)

def link_get_stats(index: "int") -> "SM_LinkStats *":
    r"""
    Get link statistics.

    Parameters
    ----------
    index: int
        The link index.

    Returns
    -------
    linkStats: SM_LinkStats *
        The link Stats struct (see :ref: SM_LinkStats). pre-allocated by the caller.

    """
    return _solver.link_get_stats(index)

def pump_get_stats(index: "int") -> "SM_PumpStats *":
    r"""
    Get pump statistics.

    Parameters
    ----------
    index: int
        The index of a pump

    Returns
    -------
    pumpStats: SM_PumpStats *
        The link Stats struct (see :ref: SM_PumpStats). pre-allocated by the caller.

    """
    return _solver.pump_get_stats(index)

def subcatch_get_stats(index: "int") -> "SM_SubcatchStats *":
    r"""
    Get subcatchment statistics.

    Parameters
    ----------
    index: int
        The index of a subcatchment

    Returns
    -------
    subcatchStats: SM_SubcatchStats *
        The link Stats struct (see :ref: SM_SubcatchStats).
        pre-allocated by the caller. Caller is also responsible for freeing the
        SM_SubcatchStats structure using swmm_freeSubcatchStats(). This frees any
        pollutants array.

    """
    return _solver.subcatch_get_stats(index)

def system_get_routing_totals() -> "SM_RoutingTotals *":
    r"""
    Get system routing totals.

    Returns
    -------
    routingTot: SM_RoutingTotals *
        The system Routing Stats struct (see :ref: SM_RoutingTotals).
        pre-allocated by the caller.

    """
    return _solver.system_get_routing_totals()

def system_get_runoff_totals() -> "SM_RunoffTotals *":
    r"""
    Get system runoff totals.

    Returns
    -------
    runoffTot: SM_RunoffTotals *
        The system Runoff Stats struct (see :ref: SM_RunoffTotals).
        pre-allocated by the caller.

    """
    return _solver.system_get_runoff_totals()

def link_set_target_setting(index: "int", setting: "double") -> "int":
    r"""
    Set a link setting (pump, orifice, or weir). Setting for an orifice and a 
    weir should be [0, 1]. A setting for a pump can range from [0, inf). However, 
    if a pump is set to 1, it will pump at its maximum curve setting.

    Parameters
    ----------
    index: int
        The link index.
    setting: double
        The new setting for the link.

    """
    return _solver.link_set_target_setting(index, setting)

def node_set_total_inflow(index: "int", flowrate: "double") -> "int":
    r"""
    Set an inflow rate to a node. The inflow rate is held constant until the 
    caller changes it.

    Parameters
    ----------
    index: int
        The node index.
    flowrate: double
        The new node inflow rate.

    """
    return _solver.node_set_total_inflow(index, flowrate)

def outfall_set_stage(index: "int", stage: "double") -> "int":
    r"""
    Set outfall stage.

    Parameters
    ----------
    index: int
        The outfall node index.
    stage: double
        The outfall node stage (head).

    """
    return _solver.outfall_set_stage(index, stage)

def raingage_set_precipitation(index: "int", total_precip: "double") -> "int":
    r"""
    Set a total precipitation intensity to the gage.

    Parameters
    ----------
    index: int
        The gage index.
    total_precip: double
        The new total precipitation intensity.

    """
    return _solver.raingage_set_precipitation(index, total_precip)

cvar = _solver.cvar




#
#  test_toolkit.py
#
#  Created:    Aug  8, 2018
#  Updated:    Apr 20, 2020
#
#  Author:     See AUTHORS
#


import os

import pytest

from swmm.toolkit import solver, shared_enum

DATA_PATH = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'data')
INPUT_FILE_EXAMPLE_1 = os.path.join(DATA_PATH, 'test_Example1.inp')
REPORT_FILE_TEST_1 = os.path.join(DATA_PATH, 'temp_Example.rpt')
OUTPUT_FILE_TEST_1 = os.path.join(DATA_PATH, 'temp_Example.out')
INPUT_FILE_EXAMPLE_2 = os.path.join(DATA_PATH, 'test_Example2.inp')
REPORT_FILE_TEST_2 = os.path.join(DATA_PATH, 'temp_Example2.rpt')
OUTPUT_FILE_TEST_2 = os.path.join(DATA_PATH, 'temp_Example2.out')
INPUT_FILE_EXAMPLE_3 = os.path.join(DATA_PATH, 'test_Example3.inp')
REPORT_FILE_TEST_3 = os.path.join(DATA_PATH, 'temp_Example3.rpt')
OUTPUT_FILE_TEST_3 = os.path.join(DATA_PATH, 'temp_Example3.out')

INPUT_FILE_FAIL = os.path.join(DATA_PATH, 'temp_nodata.inp')


def test_run():
    solver.swmm_run(INPUT_FILE_EXAMPLE_1, REPORT_FILE_TEST_1, OUTPUT_FILE_TEST_1)


def test_openclose():
    solver.swmm_open(INPUT_FILE_EXAMPLE_1, REPORT_FILE_TEST_1, OUTPUT_FILE_TEST_1)
    solver.swmm_close()


def test_errorhandling():
    with pytest.raises(Exception):
        solver.swmm_open(INPUT_FILE_FAIL, REPORT_FILE_TEST_1, OUTPUT_FILE_TEST_1)


@pytest.fixture()
def handle(request):
    solver.swmm_open(INPUT_FILE_EXAMPLE_1, REPORT_FILE_TEST_1, OUTPUT_FILE_TEST_1)

    def close():
        solver.swmm_close()

    request.addfinalizer(close)


@pytest.fixture()
def run_sim(request):
    solver.swmm_open(INPUT_FILE_EXAMPLE_1, REPORT_FILE_TEST_1, OUTPUT_FILE_TEST_1)
    solver.swmm_start(0)

    def close():
        solver.swmm_end()
        solver.swmm_close()

    request.addfinalizer(close)


@pytest.fixture()
def lid_handle(request):
    solver.swmm_open(INPUT_FILE_EXAMPLE_2, REPORT_FILE_TEST_2, OUTPUT_FILE_TEST_2)


    def close():
        solver.swmm_close()

    request.addfinalizer(close)


@pytest.fixture()
def run_lid_sim(request):
    solver.swmm_open(INPUT_FILE_EXAMPLE_2, REPORT_FILE_TEST_2, OUTPUT_FILE_TEST_2)
    solver.swmm_start(0)

    def close():
        solver.swmm_end()
        solver.swmm_close()

    request.addfinalizer(close)

@pytest.fixture()
def run_pollut_sim(request):
    solver.swmm_open(INPUT_FILE_EXAMPLE_3, REPORT_FILE_TEST_3, OUTPUT_FILE_TEST_3)
    solver.swmm_start(0)

    def close():
        solver.swmm_end()
        solver.swmm_close()

    request.addfinalizer(close)


def test_step(handle):
    solver.swmm_start(0)

    while True:
        time = solver.swmm_step()
        if time == 0:
            break

    solver.swmm_end()
    solver.swmm_report()


def test_version(handle):
    major, minor, patch = solver.swmm_version_info().split('.')
    print(major, minor, patch)
    assert major == '5'

def test_legacy_version(handle):
    version = solver.swmm_get_version()
    major = str(version)[0]
    assert isinstance(version, int)
    assert major == '5'

def test_simulation_unit(handle):
    simulation_system_unit_option = solver.simulation_get_unit(shared_enum.UnitProperty.SYSTEM_UNIT)
    simulation_flow_unit_option = solver.simulation_get_unit(shared_enum.UnitProperty.FLOW_UNIT)
    assert simulation_system_unit_option == shared_enum.UnitSystem.US.value
    assert simulation_flow_unit_option == shared_enum.FlowUnits.CFS.value


def test_object_find(handle):
    rg_index = solver.project_get_index(shared_enum.ObjectType.GAGE, 'RG1')
    sub_index_0 = solver.project_get_index(shared_enum.ObjectType.SUBCATCH, '2')
    sub_index_1 = solver.project_get_index(shared_enum.ObjectType.SUBCATCH, '3')
    node_index = solver.project_get_index(shared_enum.ObjectType.NODE, '24')
    link_index = solver.project_get_index(shared_enum.ObjectType.LINK, '16')
    assert rg_index == 0
    assert sub_index_0 == 0
    assert sub_index_1 == 1
    assert node_index == 12
    assert link_index == 12


def test_error_message_handling(handle):
    with pytest.raises(Exception) as exc_info:
        solver.project_get_index(shared_enum.ObjectType.SUBCATCH, 'sloth')

    assert 'API Key Error: Object index out of Bounds' in str(exc_info.value)


def test_object_count(handle):
    rg_count = solver.project_get_count(shared_enum.ObjectType.GAGE)
    sub_count = solver.project_get_count(shared_enum.ObjectType.SUBCATCH)
    assert rg_count == 1
    assert sub_count == 8


#def test_api_error_message(handle):
#    error_message_501 = solver.error_get_message(501)
#    assert error_message_501 == '\n API Key Error: Object Type Outside Bonds'


def test_simulation_datetime(handle):
    from datetime import datetime

    year, month, day, hour, minute, second = solver.simulation_get_datetime(shared_enum.TimeProperty.START_DATE)
    assert datetime(1998, 1, 1) == datetime(year, month, day, hour, minute, second)
    year, month, day, hour, minute, second = solver.simulation_get_datetime(shared_enum.TimeProperty.END_DATE)
    assert datetime(1998, 1, 2, 12) == datetime(year, month, day, hour, minute, second)
    year, month, day, hour, minute, second = solver.simulation_get_datetime(shared_enum.TimeProperty.REPORT_DATE)
    assert datetime(1998, 1, 1) == datetime(year, month, day, hour, minute, second)

    solver.simulation_set_datetime(shared_enum.TimeProperty.START_DATE, 2020, 1, 1, 12, 59, 59)
    year, month, day, hour, minute, second = solver.simulation_get_datetime(shared_enum.TimeProperty.START_DATE)
    assert datetime(2020, 1, 1, 12, 59, 59) == datetime(year, month, day, hour, minute, second)


def test_simulation_get_current_datetime(run_sim):
    from datetime import datetime
    current_datetime = datetime(*solver.simulation_get_current_datetime())
    assert current_datetime == datetime(1998, 1, 1)


def test_simulation_analysis_setting(handle):
    allow_ponding = solver.simulation_get_setting(0)
    skip_steady = solver.simulation_get_setting(1)
    ignore_rain = solver.simulation_get_setting(2)
    ignore_gw = solver.simulation_get_setting(5)
    assert allow_ponding == 0
    assert skip_steady == 0
    assert ignore_rain == 0
    assert ignore_gw == 1


def test_simulation_parameter(handle):
    route_step = solver.simulation_get_parameter(0)
    min_route_step = solver.simulation_get_parameter(1)
    min_slope = solver.simulation_get_parameter(6)
    head_tolerance = solver.simulation_get_parameter(11)
    assert route_step == 60
    assert min_route_step == 0.5
    assert min_slope == 0
    assert head_tolerance ==  0


def test_object_get_index(handle):
    with pytest.raises(Exception) as exc_info:
        node_index = solver.project_get_index(shared_enum.ObjectType.NODE, '6')
        sub_index = solver.project_get_index(shared_enum.ObjectType.SUBCATCH, '20')

#    assert node_index == -1
#    assert sub_index == -1

    node_index = solver.project_get_index(shared_enum.ObjectType.NODE, '20')
    sub_index = solver.project_get_index(shared_enum.ObjectType.SUBCATCH, '6')

    assert node_index == 8
    assert sub_index == 5


def test_object_get_id(handle):
    node_name = solver.project_get_id(shared_enum.ObjectType.NODE, 8)
    sub_name = solver.project_get_id(shared_enum.ObjectType.SUBCATCH, 5)
    assert node_name == '20'
    assert sub_name == '6'


def test_link_get_type(handle):
    link_type = solver.link_get_type(2)
    assert link_type == shared_enum.LinkType.CONDUIT


def test_link_get_connections(handle):
    us_node, ds_node = solver.link_get_connections(4)
    assert us_node == 9
    assert ds_node == 10

    us_node, ds_node = solver.link_get_connections(0)
    assert us_node == 0
    assert ds_node == 1


def test_link_get_direction(handle):
    for index in range(0, 13):
        link_direction = solver.link_get_direction(index)
        assert link_direction == shared_enum.LinkDirection.UPSTREAM_TO_DOWNSTREAM.value


def test_link_param(handle):
    offset_1 = solver.link_get_parameter(0, shared_enum.LinkProperty.OFFSET_1)
    offset_2 = solver.link_get_parameter(0, shared_enum.LinkProperty.OFFSET_2)
    initial_flow = solver.link_get_parameter(0, shared_enum.LinkProperty.INITIAL_FLOW)
    flow_limit = solver.link_get_parameter(0, shared_enum.LinkProperty.FLOW_LIMIT)
    inlet_loss = solver.link_get_parameter(0, shared_enum.LinkProperty.INLET_LOSS)
    outlet_loss = solver.link_get_parameter(0, shared_enum.LinkProperty.OUTLET_LOSS)
    average_loss = solver.link_get_parameter(0, shared_enum.LinkProperty.AVERAGE_LOSS)

    assert offset_1 == 0
    assert offset_2 == 0
    assert initial_flow == 0
    assert flow_limit == 0
    assert inlet_loss == 0
    assert outlet_loss == 0
    assert average_loss == 0

    solver.link_set_parameter(0, shared_enum.LinkProperty.OFFSET_1, 0.2)
    solver.link_set_parameter(0, shared_enum.LinkProperty.OFFSET_2, 0.1)
    solver.link_set_parameter(0, shared_enum.LinkProperty.INITIAL_FLOW, 0.1)
    solver.link_set_parameter(0, shared_enum.LinkProperty.FLOW_LIMIT, 1)
    solver.link_set_parameter(0, shared_enum.LinkProperty.INLET_LOSS, 0.5)
    solver.link_set_parameter(0, shared_enum.LinkProperty.OUTLET_LOSS, 1.0)
    solver.link_set_parameter(0, shared_enum.LinkProperty.AVERAGE_LOSS, 0.2)

    offset_1 = solver.link_get_parameter(0, shared_enum.LinkProperty.OFFSET_1)
    offset_2 = solver.link_get_parameter(0, shared_enum.LinkProperty.OFFSET_2)
    initial_flow = solver.link_get_parameter(0, shared_enum.LinkProperty.INITIAL_FLOW)
    flow_limit = solver.link_get_parameter(0, shared_enum.LinkProperty.FLOW_LIMIT)
    inlet_loss = solver.link_get_parameter(0, shared_enum.LinkProperty.INLET_LOSS)
    outlet_loss = solver.link_get_parameter(0, shared_enum.LinkProperty.OUTLET_LOSS)
    average_loss = solver.link_get_parameter(0, shared_enum.LinkProperty.AVERAGE_LOSS)

    assert offset_1 == 0.2
    assert offset_2 == 0.1
    assert initial_flow == 0.1
    assert flow_limit == 1
    assert inlet_loss == 0.5
    assert outlet_loss == 1.0
    assert average_loss == 0.2


def test_link_get_result(run_sim):
    flow = solver.link_get_result(0, shared_enum.LinkResult.FLOW)
    depth = solver.link_get_result(0, shared_enum.LinkResult.DEPTH)
    assert flow == 0.0
    assert depth == 0.0
    for i in range(0, 100):
        solver.swmm_step()
    flow = solver.link_get_result(0, shared_enum.LinkResult.FLOW)
    depth = solver.link_get_result(0, shared_enum.LinkResult.DEPTH)
    assert flow == pytest.approx(1.054, 0.1)
    assert depth == pytest.approx(0.268, 0.1)


def test_link_get_pollutant(run_sim):
    pollut = solver.link_get_pollutant(0, shared_enum.NodePollutant.QUALITY)
    tss, lead = tuple(pollut)
    assert tss == 0.0
    assert lead == 0.0

    for i in range(0, 250):
        solver.swmm_step()
    pollut = solver.link_get_pollutant(0, shared_enum.NodePollutant.QUALITY)

    tss, lead = tuple(pollut)

    assert tss == pytest.approx(14.7179, 0.1)
    assert lead == pytest.approx(2.9435, 0.1)

def test_link_get_pollutant_reactor(run_sim):
    pollut = solver.link_get_pollutant(0, shared_enum.LinkPollutant.REACTOR_CONC)
    tss, lead = tuple(pollut)
    assert tss == 0.0
    assert lead == 0.0

    for i in range(0, 250):
        solver.swmm_step()
    pollut = solver.link_get_pollutant(0, shared_enum.LinkPollutant.REACTOR_CONC)
    print(pollut)

    tss, lead = tuple(pollut)

    assert tss == pytest.approx(14.7179, 0.1)
    assert lead == pytest.approx(2.9435, 0.1)

def test_link_set_pollutant(run_pollut_sim):
    upstream_link = 'Valve'
    downstream_node = 'Outfall'
    pollutant = 'P1'

    upstream_link = solver.project_get_index(shared_enum.ObjectType.LINK, upstream_link)
    downstream_node = solver.project_get_index(shared_enum.ObjectType.NODE, downstream_node)
    pollutant = solver.project_get_index(shared_enum.ObjectType.POLLUT, pollutant)

    for i in range(0, 200):
        solver.link_set_pollutant(upstream_link,shared_enum.LinkPollutant.QUALITY,pollutant,1000)
        solver.swmm_step()
    assert round(solver.node_get_pollutant(downstream_node,shared_enum.NodePollutant.QUALITY)[pollutant]) == 1000

def test_link_set_target_setting(run_sim):
    target_setting = solver.link_get_result(0, shared_enum.LinkResult.TARGET_SETTING)
    assert target_setting == 1.0
    solver.link_set_target_setting(0, 0.5)
    target_setting = solver.link_get_result(0, shared_enum.LinkResult.TARGET_SETTING)
    assert target_setting == 0.5

    for i in range(0, 250):
        solver.swmm_step()

    target_setting = solver.link_get_result(0, shared_enum.LinkResult.TARGET_SETTING)
    target_setting == 0.5


def test_link_get_stats(run_sim):
    while True:
        time = solver.swmm_step()
        if time == 0:
            break

    link_stats = solver.link_get_stats(0)


def test_pump_get_stats(run_sim):
    index = solver.project_get_index(shared_enum.ObjectType.LINK, 'P1')
    while True:
        time = solver.swmm_step()
        if time == 0:
            break

    stats = solver.pump_get_stats(index)


def test_node_get_type(handle):
    node_type = solver.node_get_type(8)
    assert node_type == shared_enum.NodeType.JUNCTION
    node_index = solver.project_get_index(shared_enum.ObjectType.NODE, '18')
    node_type = solver.node_get_type(node_index)
    assert node_type == shared_enum.NodeType.OUTFALL


def test_node_param(handle):
    invert_elevation = solver.node_get_parameter(0, shared_enum.NodeProperty.INVERT_ELEVATION)
    full_depth = solver.node_get_parameter(0, shared_enum.NodeProperty.FULL_DEPTH)
    surcharge_depth = solver.node_get_parameter(0, shared_enum.NodeProperty.SURCHARGE_DEPTH)
    pond_area = solver.node_get_parameter(0, shared_enum.NodeProperty.POND_AREA)
    initial_depth = solver.node_get_parameter(0, shared_enum.NodeProperty.INITIAL_DEPTH)

    assert invert_elevation == 1000
    assert full_depth == 3
    assert surcharge_depth == 0
    assert pond_area == 0
    assert initial_depth == 0

    solver.node_set_parameter(0, shared_enum.NodeProperty.INVERT_ELEVATION, 90)
    solver.node_set_parameter(0, shared_enum.NodeProperty.FULL_DEPTH, 10)
    solver.node_set_parameter(0, shared_enum.NodeProperty.SURCHARGE_DEPTH, 100)
    solver.node_set_parameter(0, shared_enum.NodeProperty.POND_AREA, 100)
    solver.node_set_parameter(0, shared_enum.NodeProperty.INITIAL_DEPTH, 1)

    invert_elevation = solver.node_get_parameter(0, shared_enum.NodeProperty.INVERT_ELEVATION)
    full_depth = solver.node_get_parameter(0, shared_enum.NodeProperty.FULL_DEPTH)
    surcharge_depth = solver.node_get_parameter(0, shared_enum.NodeProperty.SURCHARGE_DEPTH)
    pond_area = solver.node_get_parameter(0, shared_enum.NodeProperty.POND_AREA)
    initial_depth = solver.node_get_parameter(0, shared_enum.NodeProperty.INITIAL_DEPTH)

    assert invert_elevation == 90
    assert full_depth == 10
    assert surcharge_depth == 100
    assert pond_area == 100
    assert initial_depth == 1


def test_node_get_result(run_sim):
    total_inflow = solver.node_get_result(0, shared_enum.NodeResult.TOTAL_INFLOW)
    total_outflow = solver.node_get_result(0, shared_enum.NodeResult.TOTAL_OUTFLOW)
    assert total_inflow == 0.0
    assert total_outflow == 0.0
    for i in range(0, 100):
        solver.swmm_step()
    total_inflow = solver.node_get_result(0, shared_enum.NodeResult.TOTAL_INFLOW)
    total_outflow = solver.node_get_result(0, shared_enum.NodeResult.TOTAL_OUTFLOW)
    assert total_inflow == pytest.approx(1.07, 0.1)
    assert total_outflow == pytest.approx(1.07, 0.1)


def test_node_get_pollutant(run_sim):
    pollut = solver.node_get_pollutant(0, shared_enum.NodePollutant.QUALITY)
    tss, lead = tuple(pollut)
    assert tss == 0.0
    assert lead == 0.0

    for i in range(0, 250):
        solver.swmm_step()
    pollut = solver.node_get_pollutant(0, shared_enum.NodePollutant.QUALITY)

    tss, lead = tuple(pollut)
    assert tss == pytest.approx(14.7179, 0.1)
    assert lead == pytest.approx(2.9435, 0.1)

def test_node_get_pollutant_reactor(run_pollut_sim):
    # Need to test on a storage node
    node_index = solver.project_get_index(shared_enum.ObjectType.NODE, 'Tank')
    #print(node_index)
    pollut = solver.node_get_pollutant(1, shared_enum.NodePollutant.REACTOR_CONC)
    p1 = pollut[0]
    assert p1 == 0.0

    for i in range(0, 250):
        solver.swmm_step()
        pollut = solver.node_get_pollutant(1, shared_enum.NodePollutant.REACTOR_CONC)

    p1 = pollut[0]
    assert p1 == pytest.approx(2.4455, 0.1)

def test_node_get_pollutant_inflow(run_pollut_sim):
    # Need to test on a storage node
    node_index = solver.project_get_index(shared_enum.ObjectType.NODE, 'Tank')
    #print(node_index)
    pollut = solver.node_get_pollutant(1, shared_enum.NodePollutant.INFLOW_CONC)
    p1 = pollut[0]
    assert p1 == 0.0

    for i in range(0, 250):
        solver.swmm_step()
        pollut = solver.node_get_pollutant(1, shared_enum.NodePollutant.INFLOW_CONC)

    p1 = pollut[0]
    assert p1 == pytest.approx(10.0, 0.1)

def test_node_set_pollutant(run_pollut_sim):
    upstream_node = 'Tank'
    downstream_node = 'Outfall'
    pollutant = 'P1'

    upstream_node = solver.project_get_index(shared_enum.ObjectType.NODE, upstream_node)
    downstream_node = solver.project_get_index(shared_enum.ObjectType.NODE, downstream_node)
    pollutant = solver.project_get_index(shared_enum.ObjectType.POLLUT, pollutant)

    for i in range(0, 200):
        solver.node_set_pollutant(upstream_node,shared_enum.NodePollutant.QUALITY, pollutant,1000)
        solver.swmm_step()
    assert round(solver.node_get_pollutant(downstream_node,shared_enum.NodePollutant.QUALITY)[pollutant]) == 1000

def test_node_get_hrt(run_pollut_sim):
    upstream_node = 'Tank'
    upstream_node = solver.project_get_index(shared_enum.ObjectType.NODE, upstream_node)


    for i in range(0, 200):        
        solver.swmm_step()
    assert round(solver.node_get_result(upstream_node, shared_enum.NodeResult.HYD_RES_TIME)) == 17


def test_node_total_inflow(run_sim):
    total_inflow = solver.node_get_total_inflow(0)
    assert total_inflow == 0.0

    for i in range(0, 550):
        solver.swmm_step()

    total_inflow = solver.node_get_total_inflow(0)
    assert total_inflow == pytest.approx(42639.814, 0.1)

    solver.node_set_total_inflow(0, 1.0)

    for i in range(0, 550):
        solver.swmm_step()

    total_inflow = solver.node_get_total_inflow(0)
    assert total_inflow == pytest.approx(75644.471, 0.1)


def test_outfall_set_stage(run_sim):
    head = solver.node_get_result(13, shared_enum.NodeResult.HEAD)
    head == 975

    for i in range(0, 100):
        solver.swmm_step()

    head = solver.node_get_result(13, shared_enum.NodeResult.HEAD)
    assert head == pytest.approx(975.519, 0.1)

    solver.outfall_set_stage(13, 1000)
    solver.swmm_step()


def test_subcatch_param(handle):
    width = solver.subcatch_get_parameter(0, shared_enum.SubcatchProperty.WIDTH)
    area = solver.subcatch_get_parameter(0, shared_enum.SubcatchProperty.AREA)
    impervious_fraction = solver.subcatch_get_parameter(0, shared_enum.SubcatchProperty.IMPERVIOUS_FRACTION)
    slope = solver.subcatch_get_parameter(0, shared_enum.SubcatchProperty.SLOPE)
    curb_length = solver.subcatch_get_parameter(0, shared_enum.SubcatchProperty.CURB_LENGTH)

    assert width == 500
    assert area == 10
    assert impervious_fraction == 50/100
    assert slope == 0.01/100
    assert curb_length == 0

    solver.subcatch_set_parameter(0, shared_enum.SubcatchProperty.WIDTH, 100)
    solver.subcatch_set_parameter(0, shared_enum.SubcatchProperty.AREA, 50)
    #solver.subcatch_set_parameter(0, shared_enum.SubcatchProperty.IMPERVIOUS_FRACTION, 100)
    solver.subcatch_set_parameter(0, shared_enum.SubcatchProperty.SLOPE, 0.1)
    solver.subcatch_set_parameter(0, shared_enum.SubcatchProperty.CURB_LENGTH, 10)

    width = solver.subcatch_get_parameter(0, shared_enum.SubcatchProperty.WIDTH)
    area = solver.subcatch_get_parameter(0, shared_enum.SubcatchProperty.AREA)
    #impervious_fraction = solver.subcatch_get_parameter(0, shared_enum.SubcatchProperty.IMPERVIOUS_FRACTION)
    slope = solver.subcatch_get_parameter(0, shared_enum.SubcatchProperty.SLOPE)
    curb_length = solver.subcatch_get_parameter(0, shared_enum.SubcatchProperty.CURB_LENGTH)

    assert width == 100
    assert area == 50
    #assert impervious_fraction == 100/100
    assert slope == 0.1
    assert curb_length == 10


def test_subcatch_get_connection(handle):
    object_type, object_index = solver.subcatch_get_connection(0)
    assert object_type == shared_enum.ObjectType.NODE
    assert object_index == 1

    object_type, object_index = solver.subcatch_get_connection(1)
    assert object_type == shared_enum.ObjectType.NODE
    assert object_index == 2
    object_type, object_index = solver.subcatch_get_connection(2)
    assert object_type == shared_enum.ObjectType.NODE
    assert object_index == 0
    object_type, object_index = solver.subcatch_get_connection(3)
    assert object_type == shared_enum.ObjectType.NODE
    assert object_index == 10


def test_subcatch_get_result(run_sim):
    rain = solver.subcatch_get_result(0, shared_enum.SubcatchResult.RAIN)
    evaporation = solver.subcatch_get_result(0, shared_enum.SubcatchResult.EVAPORATION)
    assert rain == 0.0
    assert evaporation == 0.0
    for i in range(0, 100):
        solver.swmm_step()
    rain = solver.subcatch_get_result(0, shared_enum.SubcatchResult.RAIN)
    evaporation = solver.subcatch_get_result(0, shared_enum.SubcatchResult.EVAPORATION)
    assert rain == pytest.approx(0.25, 0.1)
    assert evaporation == pytest.approx(0.0, 0.1)


def test_subcatch_get_pollutant(run_sim):
    pollut = solver.subcatch_get_pollutant(0, shared_enum.SubcatchPollutant.BUILD_UP)
    tss, lead = tuple(pollut)
    assert tss == 0.0
    assert lead == 0.0

    for i in range(0, 250):
        solver.swmm_step()
    pollut = solver.subcatch_get_pollutant(0, shared_enum.SubcatchPollutant.BUILD_UP)

    tss, lead = tuple(pollut)
    assert tss == pytest.approx(44.246, 0.1)
    assert lead == 0.0


def test_subcatch_get_stats(run_sim):
    while True:
        time = solver.swmm_step()
        if time == 0:
            break

    stats = solver.subcatch_get_stats(0)


def test_lid_usage_get_count(lid_handle):
    subcatch_usage_1 = solver.lid_usage_get_count(0)
    subcatch_usage_2 = solver.lid_usage_get_count(1)

    assert subcatch_usage_1 == 1
    assert subcatch_usage_2 == 0


def test_lid_usage_parameter(lid_handle):
    unit_area = solver.lid_usage_get_parameter(0, 0, shared_enum.LidUsageProperty.UNIT_AREA)
    top_width = solver.lid_usage_get_parameter(0, 0, shared_enum.LidUsageProperty.TOP_WIDTH)
    #bottom_width = solver.lid_usage_get_parameter(0, 0, shared_enum.LidUsageProperty.BOTTOM_WIDTH)
    initial_saturation = solver.lid_usage_get_parameter(0, 0, shared_enum.LidUsageProperty.INITIAL_SATURATION)
    from_impervious = solver.lid_usage_get_parameter(0, 0, shared_enum.LidUsageProperty.FROM_IMPERVIOUS)
    from_pervious = solver.lid_usage_get_parameter(0, 0, shared_enum.LidUsageProperty.FROM_PERVIOUS)

    assert unit_area == 50
    assert top_width == 10
    #assert bottom_width == pytest.approx(0)
    assert initial_saturation == 0
    assert from_impervious == 25
    assert from_pervious == 0

    solver.lid_usage_set_parameter(0, 0, shared_enum.LidUsageProperty.UNIT_AREA, 100)
    solver.lid_usage_set_parameter(0, 0, shared_enum.LidUsageProperty.TOP_WIDTH, 20)
    #solver.lid_usage_set_parameter(0, 0, shared_enum.LidUsageProperty.BOTTOM_WIDTH, 20)
    solver.lid_usage_set_parameter(0, 0, shared_enum.LidUsageProperty.INITIAL_SATURATION, 10)
    solver.lid_usage_set_parameter(0, 0, shared_enum.LidUsageProperty.FROM_IMPERVIOUS, 50)
    solver.lid_usage_set_parameter(0, 0, shared_enum.LidUsageProperty.FROM_PERVIOUS, 10)

    unit_area = solver.lid_usage_get_parameter(0, 0, shared_enum.LidUsageProperty.UNIT_AREA)
    top_width = solver.lid_usage_get_parameter(0, 0, shared_enum.LidUsageProperty.TOP_WIDTH)
    #bottom_width = solver.lid_usage_get_parameter(0, 0, shared_enum.LidUsageProperty.BOTTOM_WIDTH)
    initial_saturation = solver.lid_usage_get_parameter(0, 0, shared_enum.LidUsageProperty.INITIAL_SATURATION)
    from_impervious = solver.lid_usage_get_parameter(0, 0, shared_enum.LidUsageProperty.FROM_IMPERVIOUS)
    from_pervious = solver.lid_usage_get_parameter(0, 0, shared_enum.LidUsageProperty.FROM_PERVIOUS)

    assert unit_area == 100
    assert top_width == 20
    #assert bottom_width == 0
    assert initial_saturation == 10
    assert from_impervious == 50
    assert from_pervious == 10


def test_lid_usage_option(lid_handle):
    index = solver.lid_usage_get_option(0, 0, shared_enum.LidUsageOption.INDEX)
    number = solver.lid_usage_get_option(0, 0, shared_enum.LidUsageOption.NUMBER)
    to_perv = solver.lid_usage_get_option(0, 0, shared_enum.LidUsageOption.TO_PERV)
    drain_subcatch = solver.lid_usage_get_option(0, 0, shared_enum.LidUsageOption.DRAIN_SUBCATCH)
    drain_node = solver.lid_usage_get_option(0, 0, shared_enum.LidUsageOption.DRAIN_NODE)

    assert index == 0
    assert number == 100
    assert to_perv == 1
    assert drain_subcatch == -1
    assert drain_node == 0

    solver.lid_usage_set_option(0, 0, shared_enum.LidUsageOption.NUMBER, 200)

    number = solver.lid_usage_get_option(0, 0, shared_enum.LidUsageOption.NUMBER)
    assert number == 200


def test_lid_usage_get_flux_rate(run_lid_sim):
    while True:
        time = solver.swmm_step()
        if time == 0:
            break

    rate = solver.lid_usage_get_flux_rate(0, 0, shared_enum.LidLayer.SURFACE)


def test_lid_usage_get_result(run_lid_sim):
    while True:
        time = solver.swmm_step()
        if time == 0:
            break

    result = solver.lid_usage_get_result(0, 0, shared_enum.LidResult.SURFACE_OUTFLOW)


def test_lid_group_get_result(run_lid_sim):
    index = solver.project_get_index(shared_enum.ObjectType.SUBCATCH, 'wBC')
    while True:
        time = solver.swmm_step()
        if time == 0:
            break

    result = solver.lid_group_get_result(index, shared_enum.LidResult.OLD_DRAIN_FLOW)


def test_lid_control_get_overflow(lid_handle):
    overflow_option = solver.lid_control_get_overflow(0)
    # BC surface is greater than zero
    assert overflow_option == 0


def test_lid_control_parameter(lid_handle):
    alpha = solver.lid_control_get_parameter(0, shared_enum.LidLayer.SURFACE, shared_enum.LidLayerProperty.ALPHA)
    surface_thickness = solver.lid_control_get_parameter(0, shared_enum.LidLayer.SURFACE, shared_enum.LidLayerProperty.THICKNESS)
    soil_thickness = solver.lid_control_get_parameter(0, shared_enum.LidLayer.SOIL, shared_enum.LidLayerProperty.THICKNESS)
    storage_thickness = solver.lid_control_get_parameter(0, shared_enum.LidLayer.STORAGE, shared_enum.LidLayerProperty.THICKNESS)

    assert alpha == 1.49
    assert surface_thickness == 6
    assert soil_thickness == 12
    assert storage_thickness == 12

    solver.lid_control_set_parameter(0, shared_enum.LidLayer.SURFACE, shared_enum.LidLayerProperty.THICKNESS, 12)
    surface_thickness = solver.lid_control_get_parameter(0, shared_enum.LidLayer.SURFACE, shared_enum.LidLayerProperty.THICKNESS)
    assert surface_thickness == 12


def test_gage_precipitation(run_sim):
    total = solver.raingage_get_precipitation(0, shared_enum.RainResult.TOTAL)
    snowfall = solver.raingage_get_precipitation(0, shared_enum.RainResult.SNOWFALL)
    rainfall = solver.raingage_get_precipitation(0, shared_enum.RainResult.RAINFALL)
    assert total == 0.0
    assert snowfall == 0.0
    assert rainfall == 0.0

    for i in range(0, 250):
        solver.swmm_step()

    total = solver.raingage_get_precipitation(0, shared_enum.RainResult.TOTAL)
    snowfall = solver.raingage_get_precipitation(0, shared_enum.RainResult.SNOWFALL)
    rainfall = solver.raingage_get_precipitation(0, shared_enum.RainResult.RAINFALL)
    assert total == 0.4
    assert snowfall == 0.0
    assert rainfall == 0.4

    solver.raingage_set_precipitation(0, 1.0)

    for i in range(0, 250):
        solver.swmm_step()

    total = solver.raingage_get_precipitation(0, shared_enum.RainResult.TOTAL)
    snowfall = solver.raingage_get_precipitation(0, shared_enum.RainResult.SNOWFALL)
    rainfall = solver.raingage_get_precipitation(0, shared_enum.RainResult.RAINFALL)
    assert total == 1.0
    assert snowfall == 0.0
    assert rainfall == 1.0



# -*- coding: utf-8 -*-

#
#  __init__.py - SWMM toolkit package
#
#  Created:    Aug 9, 2018
#  Updated:    July 2, 2022
#
#  Author:     See AUTHORS
#

"""
A low level pythonic API for the swmm-output and swmm-solver dlls using SWIG.
"""


__author__ = "See AUTHORS"
__copyright__ = "None"
__credits__ = "Colleen Barr, Sam Hatchett"
__license__ = "CC0 1.0 Universal"

__version__ = "0.10.0"
__date__ = "June 7, 2021"

__maintainer__ = "Michael Tryby"
__email__ = "tryby.michael@epa.gov"
__status__  = "Beta"


import os
import platform
import sys


# Adds directory containing swmm libraries to path
if platform.system() == "Windows":
    libdir = os.path.join(os.path.dirname(__file__), "../../swmm_toolkit")

    if hasattr(os, 'add_dll_directory'):
        conda_exists = os.path.exists(os.path.join(sys.prefix, 'conda-meta'))
        if conda_exists:
            os.environ['CONDA_DLL_SEARCH_MODIFICATION_ENABLE'] = "1"
        os.add_dll_directory(libdir)
    else:
        os.environ["PATH"] = libdir + ";" + os.environ["PATH"]
